import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject
import typing as ty


class Marker(ReapyObject):

    _class_name = "Marker"
    project: reapy.Project
    project_id: int
    index: int

    def __init__(self,
                 parent_project: ty.Optional[reapy.Project] = None,
                 index: ty.Optional[int] = None,
                 parent_project_id: ty.Optional[int] = None):
        ...

    @reapy.inside_reaper()
    def _get_enum_index(self) -> int:
        """
        Return marker index as needed by RPR.EnumProjectMarkers2.
        """
        ...

    @property
    def _kwargs(self) -> ty.Dict[str, int]:
        ...

    def delete(self) -> None:
        """
        Delete marker.
        """
        ...

    @property
    def position(self) -> float:
        """
        Return marker position.

        Returns
        -------
        position : float
            Marker position in seconds.
        """
        ...

    @position.setter
    def position(self, position: float) -> None:
        """
        Set marker position.

        Parameters
        ----------
        position : float
            Marker position in seconds.
        """
        ...



import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject


class Source(ReapyObject):

    def __init__(self, id):
        self.id = id

    def __eq__(self, other):
        return isinstance(other, Source) and self.id == other.id

    @property
    def _args(self):
        return self.id,

    def delete(self):
        """
        Delete source. Be sure that no references to source remains.
        """
        RPR.PCM_Source_Destroy(self.id)

    @property
    def filename(self):
        """
        Return source file name.

        Returns
        -------
        filename : str
            Source file name.
        """
        _, filename, _ = RPR.GetMediaSourceFileName(self.id, "", 10**5)
        return filename

    @reapy.inside_reaper()
    @property
    def has_valid_id(self):
        """
        Whether ReaScript ID is still valid.

        For instance, if source has been deleted, ID will not be valid
        anymore.

        :type: bool
        """
        pointer, name = self._get_pointer_and_name()
        return any(
            RPR.ValidatePtr2(project.id, pointer, name)
            for project in reapy.get_projects()
        )

    def length(self, unit="seconds"):
        """
        Return source length in `unit`.

        Parameters
        ----------
        unit : {"beats", "seconds"}

        Returns
        -------
        length : float
            Source length in `unit`.
        """
        length, _, is_quantized = RPR.GetMediaSourceLength(self.id, 0)
        if is_quantized:
            if unit == "beats":
                return length
            elif unit == "seconds":
                raise NotImplementedError
        else:
            if unit == "beats":
                raise NotImplementedError
            elif unit == "seconds":
                return length

    @property
    def n_channels(self):
        """
        Return number of channels in source media.

        Returns
        -------
        n_channels : int
            Number of channels in source media.
        """
        n_channels = RPR.GetMediaSourceNumChannels(self.id)
        return n_channels

    @property
    def sample_rate(self):
        """
        Return source sample rate.

        Returns
        -------
        sample_rate : int
            Source sample rate.
        """
        sample_rate = RPR.GetMediaSourceSampleRate(self.id)
        return sample_rate

    @property
    def type(self):
        """
        Return source type ("WAV, "MIDI", etc.).

        Returns
        -------
        type : str
            Source type.
        """
        _, type, _ = RPR.GetMediaSourceType(self.id, "", 10**5)
        return type



import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject


class Item(ReapyObject):

    _class_name = "Item"

    def __init__(self, id):
        self.id = id

    def __eq__(self, other):
        return isinstance(other, Item) and self.id == other.id

    @property
    def _args(self):
        return self.id,

    @property
    def active_take(self):
        """
        Return the active take of the item.

        Returns
        -------
        take : Take
            Active take of the item.
        """
        take = reapy.Take(RPR.GetActiveTake(self.id))
        return take

    def add_take(self):
        """
        Create and return a new take in item.

        Returns
        -------
        take : Take
            New take in item.
        """
        take_id = RPR.AddTakeToMediaItem(self.id)
        take = reapy.Take(take_id)
        return take

    @reapy.inside_reaper()
    def delete(self):
        """Delete item."""
        RPR.DeleteTrackMediaItem(self.track.id, self.id)

    def get_info_value(self, param_name):
        value = RPR.GetMediaItemInfo_Value(self.id, param_name)
        return value

    def get_take(self, index):
        """
        Return index-th take of item.

        Parameters
        ----------
        index : int
            Take index.

        Returns
        -------
        take : Take
            index-th take of media item.
        """
        take_id = RPR.GetItemTake(self.id, index)
        take = reapy.Take(take_id)
        return take

    @reapy.inside_reaper()
    @property
    def has_valid_id(self):
        """
        Whether ReaScript ID is still valid.

        For instance, if item has been deleted, ID will not be valid
        anymore.

        :type: bool
        """
        try:
            project_id = self.project.id
        except OSError:
            return False
        pointer, name = self._get_pointer_and_name()
        return bool(RPR.ValidatePtr2(project_id, pointer, name))

    @property
    def is_selected(self):
        """
        Return whether item is selected.

        Returns
        -------
        is_selected : bool
            Whether item is selected.
        """
        is_selected = bool(RPR.IsMediaItemSelected(self.id))
        return is_selected

    @property
    def length(self):
        """
        Return item length in seconds.

        Returns
        -------
        length : float
            Item length in seconds.
        """
        param_name = "D_LENGTH"
        length = self.get_info_value(param_name)
        return length

    @length.setter
    def length(self, length):
        """
        Set item length.

        Parameters
        ----------
        length : float
            New item length in seconds.
        """
        RPR.SetMediaItemLength(self.id, length, True)

    @property
    def n_takes(self):
        """
        Return the number of takes of media item.

        Returns
        -------
        n_takes : int
            Number of takes of media item.
        """
        n_takes = RPR.GetMediaItemNumTakes(self.id)
        return n_takes

    @property
    def position(self):
        """
        Return item position in seconds.

        Returns
        -------
        position : float
            Item position in seconds.
        """
        position = self.get_info_value("D_POSITION")
        return position

    @position.setter
    def position(self, position):
        """
        Set media item position to `position`.

        Parameters
        ----------
        position : float
            New item position in seconds.
        """
        RPR.SetMediaItemPosition(self.id, position, False)

    @property
    def project(self):
        """
        Item parent project.

        :type: reapy.Project
        """
        return reapy.Project(RPR.GetItemProjectContext(self.id))

    def set_info_value(self, param_name, value):
        return RPR.SetMediaItemInfo_Value(self.id, param_name, value)

    def split(self, position):
        """
        Split item and return left and right parts.

        Parameters
        ----------
        position : float
            Split position in seconds.

        Returns
        -------
        left, right : Item
            Left and right parts of the split.
        """
        right_id = RPR.SplitMediaItem(self.id, position)
        left, right = self, Item(right_id)
        return left, right

    @reapy.inside_reaper()
    @property
    def takes(self):
        """
        Return list of all takes of media item.

        Returns
        -------
        takes : list of Take
            List of all takes of media item.
        """
        n_takes = RPR.GetMediaItemNumTakes(self.id)
        take_ids = [RPR.GetMediaItemTake(self.id, i) for i in range(n_takes)]
        takes = [reapy.Take(take_id) for take_id in take_ids]
        return takes

    @reapy.inside_reaper()
    @property
    def track(self):
        """
        Parent track of item.

        Set it by passing a track, or a track index.

        :type: Track

        Examples
        --------
        >>> track0, track1 = project.tracks[0:2]
        >>> item = track0.items[0]
        >>> item.track == track0
        True
        >>> item.track = track1  # Move to track 1
        >>> item.track = 0  # Move to track 0
        """
        track_id = RPR.GetMediaItemTrack(self.id)
        track = reapy.Track(track_id)
        return track

    @track.setter
    def track(self, track):
        if isinstance(track, int):
            track = reapy.Track(track, project=self.project)
        RPR.MoveMediaItemToTrack(self.id, track.id)

    def update(self):
        """Update item in REAPER interface."""
        RPR.UpdateItemInProject(self.id)



import functools

from reapy.errors import ExtensionNotFoundError


def depends_on_extension(extension, url):
    """Return a decorator to indicate dependency to an extension.

    If the extension is not available, an `ExtensionNotFoundError`
    will be raised when calling the decorated function.

    Parameters
    ----------
    extension : str
        Extension name.
    url : str
        URL of the download page or installation instructions of
        the extension.
    """
    message = "module 'reapy.reascript_api' has no attribute"
    def decorator(f):
        """Indicate dependency of a function to an extension.

        If the extension is not available, an `ExtensionNotFoundError`
        will be raised when calling the decorated function.
        """
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except AttributeError as exc:
                if exc.args[0].startswith(message):
                    raise ExtensionNotFoundError(extension, url)
                else:
                    raise exc
        return wrapped
    return decorator


depends_on_sws = depends_on_extension('SWS', 'www.sws-extension.org')



import builtins
import functools

import reapy


@reapy.inside_reaper()
def map(function, *iterables, constants={}, kwargs_iterable=None):
    """Efficiently map a function to iterables of arguments.

    From outside REAPER, reapy function calls are sent over the
    network for REAPER to execute them. The result in then sent back
    to the reapy client.

    When calling a large number of reapy functions, the network
    overhead can become significant. To solve this problem,
    ``reapy.map`` sends all arguments at once and retrieves all
    results at once.

    It extends the API of the built-in Python function ``map`` with
    keyword arguments options.

    It stops when the shortest of ``kwargs_iterable`` and all
    iterables in ``iterables`` is exhausted.

    Parameters
    ----------
    function : reapy callable
        Function to map to iterables of arguments.
    iterables : tuple of iterables
        Iterables of arguments that will be iterated in parallel and
        passed to ``function`` as positional arguments.
    constants : dict, optional
        Constant keyword arguments that will be passed  to
        ``function`` at each call.
    kwargs_iterable : iterable of mappings
        Iterable yielding mappings of keyword arguments that will be
        passed to ``function`` alongside the positional arguments in
        ``iterables``.

    Returns
    -------
    list
        The list of results of each calls. Unlike the built-in Python
        function ``map``, it does not return an iterator.

    Examples
    --------
    Below is a simple use case where ``reapy.map`` reduces run time
    by 95%.

    >>> import time
    >>> import reapy
    >>>
    >>> project = reapy.Project()
    >>> take = project.items[0].active_take
    >>> with reapy.inside_reaper():
    ...     start_time = time.time()
    ...     ppqs = [take.time_to_ppq(time) for time in range(10**5)]
    ...     print(f'Elapsed time without reapy.map: {time.time() - start_time:.1f} s.')
    ...
    Elapsed time without reapy.map: 15.0 s.
    >>>
    >>> start_time = time.time()
    >>> ppqs = reapy.map(take.time_to_ppq, list(range(10**5)))
    >>> print(f'Elapsed time with reapy.map: {time.time() - start_time:.1f} s.')
    Elapsed time with reapy.map: 0.7 s.
    """
    if kwargs_iterable is None:
        kwargs_iterable = iter(dict, None)
    partial_function = functools.partial(function, **constants)
    function = lambda *args: partial_function(*args[1:], **args[0])
    return list(builtins.map(function, kwargs_iterable, *iterables))



import reapy
import reapy.reascript_api as RPR
from .window import Window
import typing as ty


class MIDIEditor(Window):
    def _get_int_setting(self, setting: str) -> int:
        ...

    def _get_str_setting(self, setting: str) -> str:
        ...

    @property
    def last_clicked_cc_lane(self) -> int:
        """
        Last clicked CC lane.

        :type: int
        """
        ...

    @property
    def last_clicked_cc_lane_name(self) -> str:
        """
        Last clicked CC lane name ("velocity", "pitch", etc.).

        :type: str
        """
        ...

    @property
    def active_note_row(self) -> int:
        """
        Active note row (between 0 and 127).

        :type: int
        """
        ...

    @property
    def default_channel(self) -> int:
        """
        Default note channel (between 0 and 15).

        :type: int
        """
        ...

    @property
    def default_length(self) -> int:
        """
        Default note length in MIDI ticks.

        :type: int
        """
        ...

    @property
    def default_velocity(self) -> int:
        """
        Default note velocity (between 0 and 127).

        :type: int
        """
        ...

    @property
    def is_scale_enabled(self) -> bool:
        """
        Whether scale is enabled in editor.

        :type: bool
        """
        ...

    @property
    def is_snap_enabled(self) -> bool:
        """
        Whether snap is enabled in editor.

        :type: bool
        """
        ...

    @property
    def mode(self) -> str:
        """
        Mode of MIDI editor.

        :type: {"piano roll", "event list"}
        """
        ...

    def perform_action(self, action_id: int) -> None:
        """
        Perform action (from MIDI Editor section).

        Parameters
        ----------
        action_id : int
            Action ID.
        """
        ...

    @property
    def scale_type(self) -> str:
        """
        Scale type ID.

        :type: str
        """
        ...

    @property
    def scale_root(self) -> int:
        """
        Scale root (between 0 and 12, 0=C).

        :type: int
        """
        ...

    @property
    def take(self) -> reapy.Take:
        """
        Take currently edited.

        :type: Take
        """
        ...



import reapy
from reapy.core import ReapyObject
import reapy.reascript_api as RPR


class Window(ReapyObject):

    def __init__(self, id):
        self.id = id

    @property
    def _args(self):
        return (self.id,)

    def refresh(self):
        """Refresh window."""
        RPR.DockWindowRefreshForHWND(self.id)



import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject
import typing as ty


class Region(ReapyObject):

    _class_name = "Region"
    project_id: int
    index: int

    def __init__(self,
                 parent_project: ty.Optional[reapy.Project] = None,
                 index: ty.Optional[int] = None,
                 parent_project_id: ty.Optional[int] = None) -> None:
        ...

    def _get_enum_index(self) -> int:
        """
        Return region index as needed by RPR.EnumProjectMarkers2.
        """
        ...

    @property
    def _kwargs(self) -> ty.Dict[str, int]:
        ...

    def add_rendered_track(self, track: reapy.Track) -> None:
        """
        Add track to region render matrix for this region.

        Parameters
        ----------
        track : Track
            Track to add.

        See also
        --------
        Region.add_rendered_tracks
            Efficiently add several tracks to region render matrix.
        Region.remove_rendered_track
        Region.remove_rendered_tracks
        """
        ...

    def add_rendered_tracks(self, tracks: ty.List[reapy.Track]) -> None:
        """
        Efficiently add  several tracks to region render matrix.

        Parameters
        ----------
        tracks : list of Track
            Tracks to add.

        See also
        --------
        Region.remove_rendered_tracks
        """
        ...

    @property
    def end(self) -> float:
        """
        Region end.

        :type: float
            Region end in seconds.
        """
        ...

    @end.setter
    def end(self, end: float) -> None:
        """
        Set region end.

        Parameters
        ----------
        end : float
            region end in seconds.
        """
        ...

    def delete(self) -> None:
        """
        Delete region.
        """
        ...

    def remove_rendered_track(self, track: reapy.Track) -> None:
        """
        Remove track from region render matrix for this region.

        Parameters
        ----------
        track : Track
            Track to remove.

        See also
        --------
        Region.add_rendered_tracks
        Region.remove_rendered_track
        Region.remove_rendered_tracks
            Efficiently remove several tracks from render matrix.
        """
        ...

    def remove_rendered_tracks(self, tracks: ty.List[reapy.Track]) -> None:
        """
        Efficiently remove  several tracks from region render matrix.

        Parameters
        ----------
        tracks : list of Track
            Tracks to remove.

        See also
        --------
        Region.add_rendered_tracks
        """
        ...

    @property
    def rendered_tracks(self) -> ty.List[reapy.Track]:
        """
        List of tracks for this region in region render matrix.

        :type: list of Track
        """
        ...

    @property
    def start(self) -> float:
        """
        Region start.

        :type: float
        """
        ...

    @start.setter
    def start(self, start: float) -> None:
        """
        Set region start.

        Parameters
        ----------
        start : float
            region start in seconds.
        """
        ...



import reapy
import reapy.reascript_api as RPR
from reapy.core import ReapyObject, ReapyObjectList


class MIDIEvent(ReapyObject):

    """Abstract class for MIDI events."""

    def __init__(self, parent, index):
        """
        Create event.

        Parameters
        ----------
        parent : Take
            Take to which the event belongs.
        index : int
            Event index. It is specific to the event type, which means
            a Note and a CC can have the same index.
        """
        self.index = index
        self.parent = parent

    @property
    def _args(self):
        return self.parent, self.index

    @property
    def _del_name(self):
        return 'MIDI_DeleteEvt'

    def delete(self):
        """Delete event from the take."""
        f = getattr(RPR, self._del_name)
        f(self.parent.id, self.index)

    @reapy.inside_reaper()
    def set(self, message=None, position=None, selected=None,
            muted=None, unit="seconds", sort=True):
        """
        Set properties of event if needed.

        Parameters
        ----------
        message : Iterable[int], optional
            Can be any message buffer, for example: (0xb0, 64, 127)
            which is CC64 val127 on channel 1
        position : float, optional
            position at take
        selected : bool, optional
            Whether to select new note (default=False).
        muted : bool, optional
            Whether to mute new note (default=False).
        unit : str ("seconds as default")
            "beats"|"ppq"|"seconds" (default are seconds)
        sort : bool (True as default)
            Whether to resort notes after creating new note
            (default=True). If False, then the new note will be
            ``take.notes[-1]``. Otherwise it will be at its place in
            the time-sorted list ``take.notes``. Set to False for
            improved efficiency when adding several notes, then call
            ``Take.sort_events`` at the end.
        """
        take = self.parent
        if position:
            position = take._resolve_midi_unit((position,), unit)[0]
        if message:
            message = take._midi_to_bytestr(message)
        RPR.MIDI_SetEvt(
            take.id, self.index, selected, muted, position, message,
            len(message), not sort
        )


class MIDIEventList(ReapyObjectList):

    def __init__(self, parent):
        """
        Create event list.

        Parameters
        ----------
        parent : Take
            Take to which the event list belongs.
        """
        self.parent = parent

    def __getitem__(self, key):
        with reapy.inside_reaper():
            if key >= len(self):
                raise IndexError
            return self._elements_class(self.parent, key)

    def __len__(self):
        return getattr(self.parent, self._n_elements)

    @property
    def _args(self):
        return self.parent,

    @property
    def _elements_class(self):
        return MIDIEvent

    @property
    def _n_elements(self):
        return 'n_midi_events'


class CC(MIDIEvent):

    """MIDI CC event."""

    @property
    def channel(self):
        """
        CC channel between 0 and 15.

        :type: int

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        return RPR.MIDI_GetCC(
            self.parent.id, self.index, 0, 0, 0, 0, 0, 0, 0
        )[7]

    @property
    def channel_message(self):
        """
        CC channel message.

        :type: int

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        return RPR.MIDI_GetCC(
            self.parent.id, self.index, 0, 0, 0, 0, 0, 0, 0
        )[6]

    @property
    def _del_name(self):
        return 'MIDI_DeleteCC'

    @reapy.inside_reaper()
    @property
    def infos(self):
        """
        Return infos about CC.

        Keys are {"selected", "muted", "position", "channel",
        "channel_message", "messages"}.

        :type: dict
        """
        res = list(RPR.MIDI_GetCC(
            self.parent.id, self.index, 0, 0, 0, 0, 0, 0, 0
        ))[3:]
        res[0] = bool(res[0])
        res[1] = bool(res[1])
        res[2] = self.parent.ppq_to_time(res[2])
        res[-2] = res[-2], res[-1]
        res.pop()
        keys = (
            "selected", "muted", "position", "channel_message", "channel",
            "messages"
        )
        return {k: r for k, r in zip(keys, res)}

    @property
    def messages(self):
        """
        CC messages.

        :type: 2-list of int

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        return RPR.MIDI_GetCC(
            self.parent.id, self.index, 0, 0, 0, 0, 0, 0, 0
        )[-2:]

    @property
    def muted(self):
        """
        Whether CC is muted.

        :type: bool

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        return bool(RPR.MIDI_GetCC(
            self.parent.id, self.index, 0, 0, 0, 0, 0, 0, 0
        )[4])

    @property
    def position(self):
        """
        CC position in seconds.

        :type: float

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        return self.infos["position"]

    @property
    def selected(self):
        """
        Whether CC is selected.

        :type: bool

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        return bool(RPR.MIDI_GetCC(
            self.parent.id, self.index, 0, 0, 0, 0, 0, 0, 0
        )[3])


class CCList(MIDIEventList):

    _elements_class = CC
    _n_elements = "n_cc"


class Note(MIDIEvent):

    """MIDI note."""

    @property
    def channel(self):
        """
        Note channel between 0 and 15.

        :type: int

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return RPR.MIDI_GetNote(
            self.parent.id, self.index, 0, 0, 0, 0, 0, 0, 0
        )[7]

    @property
    def _del_name(self):
        return 'MIDI_DeleteNote'

    @property
    def end(self):
        """
        Note end in seconds.

        :type: float

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return self.infos["end"]

    @property
    def infos(self):
        """
        Return infos about note.

        Keys are {"selected", "muted", "start", "end", "channel",
        "pitch", "velocity"}.

        :type: dict
        """
        res = list(RPR.MIDI_GetNote(
            self.parent.id, self.index, 0, 0, 0, 0, 0, 0, 0
        ))[3:]
        res[0] = bool(res[0])
        res[1] = bool(res[1])
        res[2] = self.parent.ppq_to_time(res[2])
        res[3] = self.parent.ppq_to_time(res[3])
        keys = (
            "selected", "muted", "start", "end", "channel", "pitch",
            "velocity"
        )
        return {k: r for k, r in zip(keys, res)}

    @property
    def muted(self):
        """
        Whether note is muted.

        :type: bool

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return bool(RPR.MIDI_GetNote(
            self.parent.id, self.index, 0, 0, 0, 0, 0, 0, 0
        )[4])

    @property
    def pitch(self):
        """
        Note pitch between 0 and 127.

        :type: int

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return RPR.MIDI_GetNote(
            self.parent.id, self.index, 0, 0, 0, 0, 0, 0, 0
        )[8]

    @property
    def selected(self):
        """
        Whether note is selected.

        :type: bool

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return bool(RPR.MIDI_GetNote(
            self.parent.id, self.index, 0, 0, 0, 0, 0, 0, 0
        )[3])

    @property
    def start(self):
        """
        Note start in seconds.

        :type: float

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return self.infos["start"]

    @property
    def velocity(self):
        """
        Note velocity between 0 and 127.

        :type: int

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return RPR.MIDI_GetNote(
            self.parent.id, self.index, 0, 0, 0, 0, 0, 0, 0
        )[9]


class NoteList(MIDIEventList):

    _elements_class = Note
    _n_elements = "n_notes"



import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject
import typing as ty


class Take(ReapyObject):

    _class_name = "Take"
    id: str

    def __init__(self, id: str) -> None:
        ...

    def __eq__(self, other: object) -> bool:
        ...

    @property
    def _args(self) -> ty.Tuple[str]:
        ...

    def add_audio_accessor(self) -> reapy.AudioAccessor:
        """
        Create audio accessor and return it.

        Returns
        -------
        audio_accessor : AudioAccessor
            Audio accessor on take.
        """
        ...

    def add_event(self,
                  message: ty.Iterable[int],
                  position, unit: str = "seconds") -> None:
        """
        Add generic event to the take at position.

        Note
        ----
        ⋅ No sort events during this call
        ⋅ Inserting a notes within this function causes problems
            (wrong timing of notes on and offs), this is known REAPER bug.
            Use `add_note` method instead.

        Parameters
        ----------
        message : Iterable[int]
            Can be any message buffer, for example: (0xb0, 64, 127)
            which is CC64 val127 on channel 1
        position : float
            position at take
        unit : str, optional
            "beats"|"ppq"|"seconds" (default are seconds)
        """
        ...

    def add_fx(self, name: str, even_if_exists: bool = True) -> reapy.FX:
        """
        Add FX to track and return it.

        Parameters
        ----------
        name : str
            FX name.
        even_if_exists : bool, optional
            Whether the FX should be added even if there already is an
            instance of the same FX on the track (default=True).

        Returns
        -------
        fx : FX
            New FX on take (or previously existing instance of FX if
            even_if_exists=False).

        Raises
        ------
        ValueError
            If there is no FX with the specified name.
        """
        ...

    def add_note(self,
                 start: float,
                 end: float,
                 pitch: int,
                 velocity: int = 100,
                 channel: int = 0,
                 selected: bool = False,
                 muted: bool = False,
                 unit: str = "seconds",
                 sort: bool = True) -> None:
        """
        Add MIDI note to take.

        Parameters
        ----------
        start : float
            Note start. Unit depends on ``unit``.
        end : float
            Note end. Unit depends on ``unit``.
        pitch : int
            Note pitch between 0 and 127.
        velocity : int, optional
            Note velocity between 0 and 127 (default=100).
        channel : int, optional
            MIDI channel between 0 and 15.
        selected : bool, optional
            Whether to select new note (default=False).
        muted : bool
            Whether to mute new note (default=False).
        unit : {"seconds", "ppq", "beats"}, optional
            Time unit for ``start`` and ``end`` (default="seconds").
            ``"ppq"`` refers to MIDI ticks.
        sort : bool, optional
            Whether to resort notes after creating new note
            (default=True). If False, then the new note will be
            ``take.notes[-1]``. Otherwise it will be at its place in
            the time-sorted list ``take.notes``. Set to False for
            improved efficiency when adding several notes, then call
            ``Take.sort_events`` at the end.

        See also
        --------
        Take.sort_events
        """
        ...

    def add_sysex(self,
                  message: ty.Iterable[int],
                  position: float,
                  unit: str = "seconds",
                  evt_type: int = -1) -> None:
        """
        Add SysEx event to take.

        Note
        ----
        ⋅ No sort events during this call
        ⋅ No need to add 0xf0 ... 0xf7 bytes (they will be doubled)

        Parameters
        ----------
        message : Iterable[int]
            Can be any message buffer, for example: (0xb0, 64, 127)
            which is CC64 val127 on channel 1
        position : float
            position at take
        unit : str, optional
            "beats"|"ppq"|"seconds" (default are seconds)
        evt_type: int (default -1)
            Allowable types are
            ⋅ -1:sysex (msg should not include bounding F0..F7),
            ⋅ 1-14:MIDI text event types,
            ⋅ 15=REAPER notation event.
        """
        ...

    def beat_to_ppq(self, beat: float) -> float:
        """
        Convert beat number (from project start) to MIDI ticks (of the take).

        Parameters
        ----------
        beat : float
            Beat time to convert in beats.

        Returns
        -------
        ppq : float
            Converted time in MIDI ticks of current take.

        See also
        --------
        Take.ppq_to_beat
        Take.time_to_ppq
        """
        ...

    @property
    def cc_events(self) -> reapy.CCList:
        """
        List of CC events on take.

        :type: CCList
        """
        ...

    @property
    def envelopes(self) -> reapy.EnvelopeList:
        ...

    @property
    def fxs(self) -> reapy.FXList:
        """
        FXs on take.

        :type: FXList
        """
        ...

    def get_info_value(self, param_name: str) -> float:
        ...

    @property
    def has_valid_id(self) ->  bool:
        """
        Whether ReaScript ID is still valid.

        For instance, if take has been deleted, ID will not be valid
        anymore.

        :type: bool
        """

    @property
    def is_active(self) -> bool:
        """
        Whether take is active.

        :type: bool
        """
        ...

    @property
    def is_midi(self) -> bool:
        """
        Whether take contains MIDI or audio.

        :type: bool
        """
        ...

    @property
    def item(self) -> reapy.Item:
        """
        Parent item.

        :type: Item
        """
        ...

    @property
    def guid(self) -> str:
        """Used for communication within other scripts.

        :type: str
        """
        ...

    def make_active_take(self) -> None:
        """
        Make take active.
        """
        ...

    @property
    def midi_events(self) -> reapy.core.item.midi_event.MIDIEventList[
            reapy.core.item.midi_event.MIDIEvent]:
        """
        Get all midi events as EventList.

        Returns
        -------
        MIDIEventList
        """
        ...

    def midi_hash(self, notes_only: bool = False) -> str:
        """
        Get hash of MIDI-data to compare with later.

        Parameters
        ----------
        notes_only : bool, (False by default)
            count only notes if True

        Returns
        -------
        str
        """
        ...

    def _midi_to_bytestr(self, message: ty.Iterable[int]) -> str: ...

    @property
    def n_cc(self) -> int:
        """
        Number of MIDI CC events in take (always 0 for audio takes).

        :type: int
        """
        ...

    @property
    def n_envelopes(self) -> int:
        """
        Number of envelopes on take.

        :type: int
        """
        ...

    @property
    def n_fxs(self) -> int:
        """
        Number of FXs on take.

        :type: int
        """
        ...

    @property
    def n_midi_events(self) -> int:
        """
        Number of MIDI events in take.

        :type: int
        """
        ...

    @property
    def n_notes(self) -> int:
        """
        Number of MIDI notes in take (always 0 for audio takes).

        :type: int
        """
        ...

    @property
    def n_text_sysex(self) -> int:
        """
        Number of MIDI text/sysex events in take (0 for audio takes).

        :type: int
        """
        ...

    @property
    def name(self) -> str:
        """
        Take name.

        :type: str
        """
        ...

    @property
    def notes(self) -> reapy.NoteList:
        """
        List of MIDI notes on take.

        Unless ``Take.add_note`` has been called with ``sort=False``,
        notes are time-sorted.

        :type: NoteList
        """
        ...

    def ppq_to_beat(self, ppq: float) -> float:
        """
        Convert time in MIDI ticks to beats.

        Note
        ----
        ticks counted from take's start,
        beats from project's start

        Parameters
        ----------
        ppq : float
            Time to convert in MIDI ticks.

        Returns
        -------
        beat : float
            Converted time in beats.

        See also
        --------
        Take.beat_to_ppq
        Take.ppq_to_time
        """
        ...

    def ppq_to_time(self, ppq: float) -> float:
        """
        Convert time in MIDI ticks to seconds.

        Parameters
        ----------
        ppq : float
            Time to convert in MIDI ticks.

        Returns
        -------
        time : float
            Converted time in seconds.

        See also
        --------
        Take.time_to_ppq
        """
        ...

    @property
    def project(self) -> reapy.Project:
        """
        Take parent project.

        :type: reapy.Project
        """

    def _resolve_midi_unit(self,
                           pos_tuple: ty.Tuple[float, ...],
                           unit: str = "seconds") -> ty.Tuple[float, ...]:
        """Get positions as ppq from tuple of positions of any length.

        Parameters
        ----------
        pos_tuple : Tuple[float]
            tuple of position time in bets, ppq or seconds.
        unit : str, optional
            type of position inside tuple: seconds|beats|ppq

        Returns
        -------
        Tuple[float]
            the same tuple normalized to ppq
        """
        ...

    def select_all_midi_events(self, select: bool = True) -> None:
        """
        Select or unselect all MIDI events.

        Parameters
        ----------
        select : bool
            Whether to select or unselect events.

        See also
        --------
        Take.unselect_all_midi_events
        """
        ...

    def set_info_value(self, param_name: str, value: float) -> bool:
        ...

    def sort_events(self) -> None:
        """
        Sort MIDI events on take.

        This is only needed if ``Take.add_note`` was called with
        ``sort=False``.

        Examples
        --------
        The following example creates 100 MIDI notes on take in
        reversed order, with ``sort=False`` for efficiency. Thus,
        ``take.notes`` is not time-sorted. ``take.sort_events`` is
        called afterwards so that ``take.notes`` is time-sorted.

        >>> for i in range(100):
        ...     take.add_note(99 - i, 100 - i, pitch=0, sort=False)
        ...
        >>> take.notes[0].start, take.notes[1].start
        99.0, 98.0
        >>> take.sort_events()
        >>> take.notes[0].start, take.notes[1].start
        0.0, 1.0
        """
        ...

    @property
    def source(self) -> reapy.Source:
        """
        Take source.

        :type: Source
        """
        ...

    @property
    def start_offset(self) -> float:
        """
        Start time of the take relative to start of source file.

        :type: float
        """
        ...

    def time_to_ppq(self, time: float) -> float:
        """
        Convert time in seconds to MIDI ticks.

        Parameters
        ----------
        time : float
            Time to convert in seconds.

        Returns
        -------
        ppq : float
            Converted time in MIDI ticks.

        See also
        --------
        Take.ppq_to_time
        """
        ...

    @property
    def track(self) -> reapy.Track:
        """
        Parent track of take.

        :type: Track
        """
        ...

    def unselect_all_midi_events(self) -> None:
        """
        Unselect all MIDI events.

        See also
        --------
        Take.select_all_midi_events
        """
        ...

    @property
    def visible_fx(self) -> ty.Optional[reapy.FX]:
        """
        Visible FX in FX chain if any, else None.

        :type: FX or NoneType
        """
        ...



import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject


class Marker(ReapyObject):

    _class_name = "Marker"

    def __init__(
        self, parent_project=None, index=None, parent_project_id=None
    ):
        if parent_project_id is None:
            message = (
                "One of `parent_project` or `parent_project_id` must be "
                "specified."
            )
            assert parent_project is not None, message
            parent_project_id = parent_project.id
        self.project = reapy.Project(parent_project_id)
        self.project_id = parent_project_id
        if index is None:
            index = len(self.project.markers)
        self.index = index

    @reapy.inside_reaper()
    def _get_enum_index(self):
        """
        Return marker index as needed by RPR.EnumProjectMarkers2.
        """
        for index, marker in enumerate(self.project.markers):
            if marker.index == self.index:
                return index

    @property
    def _kwargs(self):
        return {
            "index": self.index, "parent_project_id": self.project_id
        }

    def delete(self):
        """
        Delete marker.
        """
        RPR.DeleteProjectMarker(self.project_id, self.index, False)

    @reapy.inside_reaper()
    @property
    def position(self):
        """
        Return marker position.

        Returns
        -------
        position : float
            Marker position in seconds.
        """
        index = self._get_enum_index()
        return RPR.EnumProjectMarkers2(self.project_id, index, 0, 0, 0, 0, 0)[4]

    @position.setter
    def position(self, position):
        """
        Set marker position.

        Parameters
        ----------
        position : float
            Marker position in seconds.
        """
        RPR.SetProjectMarker2(
            self.project_id, self.index, False, position, 0, ""
        )



from reapy.errors import DisconnectedClientError, DistError
from reapy.tools import json
from .socket import Socket
import typing as ty


class Client(Socket):
    address: str
    port: int
    host: str

    def __init__(self, port: int, host: str = 'localhost') -> None:
        ...

    def _connect(self, port: int, host: str) -> None:
        ...

    def _get_result(self) -> ty.Any:
        ...

    def request(self,
                function: ty.Callable[..., ty.Any],
                input: ty.Optional[object] = None) -> ty.Any:
        ...



import reapy
from reapy.core import ReapyObject
import reapy.reascript_api as RPR
import typing as ty


class Window(ReapyObject):
    id: bytes

    def __init__(self, id: bytes) -> None:
        ...

    @property
    def _args(self) -> ty.Tuple[bytes]:
        ...

    def refresh(self) -> None:
        """Refresh window."""
        ...



import reapy
import reapy.reascript_api as RPR
from reapy.core import ReapyObject


class AudioAccessor(ReapyObject):

    def __init__(self, id):
        self.id = id

    @property
    def _args(self):
        return self.id,

    def delete(self):
        """Delete audio accessor."""
        RPR.DestroyAudioAccessor(self.id)

    @property
    def end_time(self):
        """
        End time of audio that can be returned from this accessor.

        Return value is in seconds.

        :type: float
        """
        return RPR.GetAudioAccessorEndTime(self.id)

    def get_samples(
        self, start, n_samples_per_channel, n_channels=1, sample_rate=44100
     ):
        """
        Return audio samples.

        Parameters
        ----------
        start : float
            Start time in seconds.
        n_samples_per_channel : int
            Number of required samples per channel
        n_channels : int, optional
            Number of required channels (default=1).
        sample_rate : float, optional
            Required sample rate (default=44100).

        Returns
        -------
        samples : list
            List of length n_samples*n_channels.

        Examples
        --------
        To separate channels use:
        
        >>> samples = audio_accessor.get_samples(0, 1024, 2)
        >>> first_channel = samples[::2]
        >>> second_channel = samples[1::2]
        """
        buffer = [0]*n_channels*n_samples_per_channel
        samples = RPR.GetAudioAccessorSamples(
            self.id, sample_rate, n_channels, start, n_samples_per_channel,
            buffer
        )[1]
        return samples

    @property
    def has_state_changed(self):
        """
        Whether underlying state has changed.

        :type: bool
        """
        return bool(RPR.AudioAccessorValidateState(self.id))

    def hash(self):
        """
        String that changes only if the underlying samples change.

        :type: str
        """
        return RPR.GetAudioAccessorHash(self.id, "")[1]

    @property
    def start_time(self):
        """
        Start time of audio that can be returned from this accessor.

        Return value is in seconds.

        :type: float
        """
        return RPR.GetAudioAccessorStartTime(self.id)



"""Audio handling functions."""

import reapy
import reapy.reascript_api as RPR


def get_input_latency(unit="second"):
    """
    Return input latency.

    Parameters
    ----------
    unit : {"sample", "second"}
        Whether to return latency in samples or seconds
        (default="second").

    Returns
    -------
    latency : float
        Input latency.
    """
    latency, out_latency = RPR.GetInputOutputLatency(0, 0)
    if unit == "second":
        # Small hack because RPR.GetInputLatency doesn't exist...
        latency *= RPR.GetOutputLatency()/out_latency
    return latency


@reapy.inside_reaper()
def get_input_names():
    """
    Return names of all input channels.

    Returns
    -------
    names : list of str
        Names of input channels.
    """
    n_channels = reapy.audio.get_n_inputs()
    return tuple(map(RPR.GetInputChannelName, range(n_channels)))


def get_n_inputs():
    """
    Return number of audio inputs.

    Returns
    -------
    n_inputs : int
        Number of audio inputs.
    """
    n_inputs = RPR.GetNumAudioInputs()
    return n_inputs


def get_n_outputs():
    """
    Return number of audio outputs.

    Returns
    -------
    n_outputs : int
        Number of audio outputs.
    """
    n_outputs = RPR.GetNumAudioOutputs()
    return n_outputs


def get_output_latency(unit="second"):
    """
    Return output latency.

    Parameters
    ----------
    unit : {"sample", "second"}
        Whether to return latency in samples or seconds
        (default="second").

    Returns
    -------
    latency : float
        Output latency.
    """
    latency, out_latency = RPR.GetInputOutputLatency(0, 0)
    if unit == "second":
        latency = RPR.GetOutputLatency()
    else:
        latency = RPR.GetInputOutputLatency(0, 0)[1]
    return latency


@reapy.inside_reaper()
def get_output_names():
    """
    Return names of all output channels.

    Returns
    -------
    names : list of str
        Names of output channels.
    """
    n_channels = reapy.audio.get_n_outputs()
    return tuple(map(RPR.GetOutputChannelName, range(n_channels)))


def init():
    """
    Open all audio and MIDI devices (if not opened).
    """
    RPR.Audio_Init()


def is_prebuffer():
    """
    Return whether audio is in pre-buffer (threadsafe).

    Returns
    -------
    is_prebuffer : bool
        Whether audio is in pre-buffer.
    """
    is_prebuffer = bool(RPR.Audio_IsPreBuffer())
    return is_prebuffer


def is_running():
    """
    Return whether audio is running (threadsafe).

    Returns
    -------
    is_running : bool
        Whether audio is running.
    """
    is_running = bool(RPR.Audio_IsRunning())
    return is_running


def quit():
    """
    Close all audio and MIDI devices (if opened).
    """
    RPR.Audio_Quit()



"""Get setup infos."""

from reapy.reascripts import enable_dist_api, disable_dist_api

import os
import sys
import typing as ty


def get_config_scripts() -> ty.Tuple[str, str]:
    """
    Return paths to configuration ReaScripts.
    """
    ...


def get_python_dll() -> str:
    """
    Return path to Python DLL (if it can be found).
    """
    ...


string: str



import functools
import typing as ty

from reapy.errors import ExtensionNotFoundError

FuncType = ty.Callable[..., ty.Any]
F = ty.TypeVar('F', bound=FuncType)


def depends_on_extension(extension: str, url: str) -> ty.Callable[[F], F]:
    """Return a decorator to indicate dependency to an extension.

    If the extension is not available, an `ExtensionNotFoundError`
    will be raised when calling the decorated function.

    Parameters
    ----------
    extension : str
        Extension name.
    url : str
        URL of the download page or installation instructions of
        the extension.
    """
    ...


depends_on_sws = depends_on_extension('SWS', 'www.sws-extension.org')



import reapy.reascript_api as RPR
from .window import Window


class ToolTip(Window):

    """Tooltip window."""

    def __init__(self, message=" ", x=0, y=0, topmost=True, show=True):
        """Initialize tooltip.

        Parameters
        ----------
        message : str, optional
            ToolTip message (default=" "). Note that tooltips with
            empty messages are always hidden.
        x : int, optional
            x position (default=0).
        y : int, optional
            y position (default=0).
        topmost : bool, optional
            Whether tooltip should be displayed on top of all other
            windows (default=True).
        show : bool, optional
            Whether to show tooltip on initialization (default=True).
        """
        self._message = message
        self._x = x
        self._y = y
        self._topmost = topmost
        if show:
            self.show()
        self.id = RPR.GetTooltipWindow

    def hide(self):
        """Hide tooltip."""
        RPR.TrackCtl_SetToolTip("", self.x, self.y, self.topmost)
        self._is_shown = False

    @property
    def message(self):
        """
        Tooltip message.

        Note that tooltips with empty messages are always hidden.

        :type: str
        """
        return self._message

    @message.setter
    def message(self, message):
        self._message = message
        if self._is_shown:
            self.show()

    def refresh(self):
        raise NotImplementedError

    def show(self):
        """Show tooltip."""
        RPR.TrackCtl_SetToolTip(self.message, self.x, self.y, self.topmost)
        self._is_shown = True

    @property
    def topmost(self):
        """
        Whether tooltip is displayed on top of all other windows.

        :type: bool
        """
        return self._topmost

    @topmost.setter
    def topmost(self, topmost):
        self._topmost = topmost
        if self._is_shown:
            self.show()

    @property
    def x(self):
        """
        x position.

        :type: int"""
        return self._x

    @x.setter
    def x(self, x):
        self._x = x
        if self._is_shown:
            self.show()

    @property
    def y(self):
        """y position.

        :type: int
        """
        return self._y

    @y.setter
    def y(self, y):
        self._y = y
        if self._is_shown:
            self.show()



"""Define reapy.defer and reapy.at_exit."""

import reapy

import os
import sys
import tempfile
import typing as ty


class Deferrer:
    """Class to register and run deferred calls."""

    _next_call_id: int
    _callbacks: ty.Dict[int, ty.Callable[..., ty.Any]]
    _args: ty.Dict[int, ty.Tuple[ty.Any, ...]]
    _kwargs: ty.Dict[int, ty.Dict[str, ty.Any]]

    def _get_new_call_id(self) -> int:
        ...

    def _wrapped_open(self, *args: ty.Any,
                      **kwargs: ty.Any) -> ty.Union[ReaperConsole, ty.Any]:
        ...

    def defer(self,
              callback: ty.Callable[..., ty.Any],
              args: ty.Tuple[ty.Any, ...],
              kwargs: ty.Dict[str, ty.Any],
              at_exit: bool = False) -> None:
        ...

    def run(self, call_id: int) -> None:
        ...


class ReaperConsole:
    """File-like wrapper around the Reaper Console."""
    def close(self) -> None:
        ...

    def flush(self) -> None:
        ...

    def write(self, *args: ty.Any, **kwargs: ty.Any) -> None:
        ...


def at_exit(f: ty.Callable[..., ty.Any], *args: ty.Any,
            **kwargs: ty.Any) -> None:
    """
    Make REAPER call a function after script execution.

    The function is also called if excution is terminated by user.

    Parameters
    ----------
    f : callable
        Function to be called later.
    args : tuple, optional
        Positional arguments to pass to ``f``.
    kwargs : dict, optional
        Keyword arguments to pass to ``f``.

    Raises
    ------
    AssertionError
        When called from outside REAPER.

    Examples
    --------
    Typical use case of ``at_exit`` is cleaning up after a ``defer``
    loop.

    The following example opens a file and starts a loop that
    indefinitely writes integers to that file. Since we want the file
    to be closed when the user terminates script execution, call to
    its ``close`` method is deferred to ``reapy.at_exit``.

    >>> import reapy
    >>> file = open("somefile.txt", "w")
    >>> def stupid_loop(i):
    ...     file.write(i)
    ...     reapy.defer(stupid_loop, i + 1)
    ...
    >>> reapy.at_exit(file.close)
    >>> stupid_loop(0)
    """
    ...


def defer(f: ty.Callable[..., ty.Any], *args: ty.Any,
          **kwargs: ty.Any) -> None:
    """
    Make REAPER call a function later.

    Parameters
    ----------
    f : callable
        Function to be called later.
    args : tuple, optional
        Positional arguments to pass to ``f``.
    kwargs : dict, optional
        Keyword arguments to pass to ``f``.

    Raises
    ------
    AssertionError
        When called from outside REAPER.

    Notes
    -----
    The average time before a defered call is actually run is about
    0.03 seconds (around 30 defered calls are allowed per second).

    Examples
    --------
    Typical use case of ``defer`` is running loops that don't block
    REAPER GUI.

    The following example creates a loop that indefinitely prints
    integers to the REAPER console, without blocking REAPER GUI.

    >>> import reapy
    >>> def stupid_loop(i):
    ...     reapy.print(i)
    ...     reapy.defer(stupid_loop, i + 1)
    ...
    >>> stupid_loop(0)
    """
    # Check we are inside REAPER
    ...



"""Defines class Project."""

import pickle
import codecs
import os

import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject
from reapy.errors import RedoError, UndoError


class Project(ReapyObject):

    """REAPER project."""

    def __init__(self, id=None, index=-1):
        """
        Build project either by ID or index.

        Parameters
        ----------
        id : None, str or int, optional
            Project identifier.
            When None (default), `index is used instead.
            An integer is interpreted as the project index in GUI.
            A string starting with '(ReaProject*)0x' is interpreted
            as a ReaScript identifier.
            Otherwise, `id` is the project name. In that case, the .rpp
            extension is optional.
        index : int, optional
            Project index in GUI (default=-1, corresponds to current
            project).
        """
        if isinstance(id, int):
            id, index = None, id
        if id is None:
            id = RPR.EnumProjects(index, None, 0)[0]
        if not id.startswith('(ReaProject*)0x'):
            id = Project._from_name(id).id
        self.id = id
        self._filename = None

    def __eq__(self, other):
        if hasattr(other, 'id'):
            return self.id == other.id
        return False

    @property
    def _args(self):
        return self.id,

    @staticmethod
    def _from_name(name):
        """Return project with corresponding name.

        Parameters
        ----------
        name : str
            Project file name. Including the extension ('.rpp')
            is optional.

        Returns
        -------
        Project

        Raises
        ------
        NameError
            If no project with the corresponding name is open.
        """
        if not name.lower().endswith('.rpp'):
            name += '.rpp'
        with reapy.inside_reaper():
            for project in reapy.get_projects():
                project_name = project.name[:-4] + '.rpp'
                if project_name == name:
                    return project
        raise NameError('"{}" is not currently open.'.format(name))

    @reapy.inside_reaper()
    def _get_track_by_name(self, name):
        """Return first track with matching name."""
        for track in self.tracks:
            if track.name == name:
                return track
        raise KeyError(name)

    def add_marker(self, position, name="", color=0):
        """
        Create new marker and return its index.

        Parameters
        ----------
        position : float
            Marker position in seconds.
        name : str, optional
            Marker name.
        color : int or tuple of int, optional
            Marker color. Integers correspond to REAPER native colors.
            Tuple must be RGB triplets of integers between 0 and 255.

        Returns
        -------
        marker : reapy.Marker
            New marker.

        Notes
        -----
        If a marker with the same position and name already exists, no
        new marker will be created, and the existing marker index will
        be returned.
        """
        if isinstance(color, tuple):
            color = reapy.rgb_to_native(color) | 0x1000000
        marker_id = RPR.AddProjectMarker2(
            self.id, False, position, 0, name, -1, color
        )
        marker = reapy.Marker(self, marker_id)
        return marker

    def add_region(self, start, end, name="", color=0):
        """
        Create new region and return its index.

        Parameters
        ----------
        start : float
            Region start in seconds.
        end : float
            Region end in seconds.
        name : str, optional
            Region name.
        color : int or tuple of int, optional
            Region color. Integers correspond to REAPER native colors.
            Tuple must be RGB triplets of integers between 0 and 255.

        Returns
        -------
        region : reapy.Region
            New region.
        """
        if isinstance(color, tuple):
            color = reapy.rgb_to_native(color) | 0x1000000
        region_id = RPR.AddProjectMarker2(
            self.id, True, start, end, name, -1, color
        )
        region = reapy.Region(self, region_id)
        return region

    @reapy.inside_reaper()
    def add_track(self, index=0, name=""):
        """
        Add track at a specified index.

        Parameters
        ----------
        index : int
            Index at which to insert track.
        name : str, optional
            Name of created track.

        Returns
        -------
        track : Track
            New track.
        """
        n_tracks = self.n_tracks
        index = max(-n_tracks, min(index, n_tracks))
        if index < 0:
            index = index % n_tracks
        with self.make_current_project():
            RPR.InsertTrackAtIndex(index, True)
        track = self.tracks[index]
        track.name = name
        return track

    @property
    def any_track_solo(self):
        """
        Test whether any track is soloed in project.

        Returns
        -------
        any_track_solo : bool
            Whether any track is soloed in project.
        """
        any_track_solo = bool(RPR.AnyTrackSolo(self.id))
        return any_track_solo

    def beats_to_time(self, beats):
        """
        Convert beats to time in seconds.

        Parameters
        ----------
        beats : float
            Time in beats

        Returns
        -------
        time : float
            Converted time in seconds.

        See also
        --------
        Project.time_to_beats
        """
        time = RPR.TimeMap2_QNToTime(self.id, beats)
        return time

    def begin_undo_block(self):
        """
        Start a new undo block.
        """
        RPR.Undo_BeginBlock2(self.id)

    @property
    def bpi(self):
        """
        Return project BPI (numerator of time signature).

        Returns
        -------
        bpi : float
            Numerator of time signature.
        """
        return self.time_signature[1]

    @reapy.inside_reaper()
    @property
    def bpm(self):
        """
        Project BPM (beats per minute).

        :type: float
        """
        return self.time_signature[0]

    @bpm.setter
    def bpm(self, bpm):
        """
        Set project BPM (beats per minute).

        Parameters
        ----------
        bpm : float
            Tempo in beats per minute.
        """
        RPR.SetCurrentBPM(self.id, bpm, True)

    @property
    def buffer_position(self):
        """
        Position of next audio block being processed in seconds.

        :type: float

        See also
        --------
        Project.play_position
            Latency-compensated actual-what-you-hear position.
        """
        return RPR.GetPlayPosition2Ex(self.id)

    @reapy.inside_reaper()
    def bypass_fx_on_all_tracks(self, bypass=True):
        """
        Bypass or un-bypass FX on all tracks.

        Parameters
        ----------
        bypass : bool
            Whether to bypass or un-bypass FX.
        """
        with self.make_current_project():
            RPR.BypassFxAllTracks(bypass)

    @property
    def can_redo(self):
        """
        Whether redo is possible.

        :type: bool
        """
        try:
            RPR.Undo_CanRedo2(self.id)
            can_redo = True
        except AttributeError:  # Bug in ReaScript API
            can_redo = False
        return can_redo

    @property
    def can_undo(self):
        """
        Whether undo is possible.

        :type: bool
        """
        try:
            RPR.Undo_CanUndo2(self.id)
            can_undo = True
        except AttributeError:  # Bug in ReaScript API
            can_undo = False
        return can_undo

    def close(self):
        """Close project and its correspondig tab."""
        self._filename = os.path.join(self.path, self.name)
        with self.make_current_project():
            reapy.perform_action(40860)

    @property
    def cursor_position(self):
        """
        Edit cursor position in seconds.

        :type: float
        """
        position = RPR.GetCursorPositionEx(self.id)
        return position

    @cursor_position.setter
    def cursor_position(self, position):
        """
        Set edit cursor position.

        Parameters
        ----------
        position : float
            New edit cursor position in seconds.
        """
        RPR.SetEditCurPos(position, True, True)

    @reapy.inside_reaper()
    def disarm_rec_on_all_tracks(self):
        """
        Disarm record on all tracks.
        """
        with self.make_current_project():
            RPR.ClearAllRecArmed()

    def end_undo_block(self, description=""):
        """
        End undo block.

        Parameters
        ----------
        description : str
            Undo block description.
        """
        RPR.Undo_EndBlock2(self.id, description, 0)

    @reapy.inside_reaper()
    @property
    def focused_fx(self):
        """
        FX that has focus if any, else None.

        :type: FX or NoneType
        """
        if not self.is_current_project:
            return
        res = RPR.GetFocusedFX(0, 0, 0)
        if not res[0]:
            return
        if res[1] == 0:
            track = self.master_track
        else:
            track = self.tracks[res[1] - 1]
        if res[0] == 1:  # Track FX
            return track.fxs[res[3]]
        # Take FX
        item = track.items[res[2]]
        take = item.takes[res[3] // 2**16]
        return take.fxs[res[3] % 2**16]

    def get_info_string(self, param_name: str) -> str:
        """
        Parameters
        ----------
        param_name : str
            MARKER_GUID:X : get the GUID (unique ID) of the marker or region
                with index X, where X is the index passed to
                EnumProjectMarkers, not necessarily the displayed number
            RECORD_PATH :
                recording directory -- may be blank or a relative path,
                to get the effective path see GetProjectPathEx()
            RENDER_FILE : render directory
            RENDER_PATTERN : render file name (may contain wildcards)
            RENDER_FORMAT : base64-encoded sink configuration
                (see project files, etc). Callers can also pass a simple
                4-byte string (non-base64-encoded), e.g. "evaw" or "l3pm",
                to use default settings for that sink type.
            RENDER_FORMAT2 : base64-encoded secondary sink configuration.
                Callers can also pass a simple 4-byte string (non-base64-encoded),
                e.g. "evaw" or "l3pm", to use default settings for
                that sink type, or "" to disable secondary render.
                Formats available on this machine:
                "wave" "aiff" "iso " "ddp " "flac" "mp3l" "oggv" "OggS"
                "FFMP" "GIF " "LCF " "wvpk"
        """
        _, _, _, result, _ = RPR.GetSetProjectInfo_String(
            self.id, param_name, 'valuestrNeedBig', False
        )
        return result

    def get_info_value(self, param_name: str) -> float:
        """
        Parameters
        ----------
        param_name : str
            RENDER_SETTINGS : &(1|2)=0:master mix, &1=stems+master mix,
                &2=stems only, &4=multichannel tracks to multichannel files,
                &8=use render matrix, &16=tracks with only mono media
                to mono files, &32=selected media items,
                &64=selected media items via master
            RENDER_BOUNDSFLAG : 0=custom time bounds, 1=entire project,
                2=time selection, 3=all project regions,
                4=selected media items, 5=selected project regions
            RENDER_CHANNELS : number of channels in rendered file
            RENDER_SRATE : sample rate of rendered file
                (or 0 for project sample rate)
            RENDER_STARTPOS : render start time when RENDER_BOUNDSFLAG=0
            RENDER_ENDPOS : render end time when RENDER_BOUNDSFLAG=0
            RENDER_TAILFLAG : apply render tail setting when rendering:
                &1=custom time bounds, &2=entire project, &4=time selection,
                &8=all project regions, &16=selected media items,
                &32=selected project regions
            RENDER_TAILMS : tail length in ms to render
                (only used if RENDER_BOUNDSFLAG and RENDER_TAILFLAG are set)
            RENDER_ADDTOPROJ : 1=add rendered files to project
            RENDER_DITHER : &1=dither, &2=noise shaping, &4=dither stems,
                &8=noise shaping on stems
            PROJECT_SRATE : samplerate (ignored unless PROJECT_SRATE_USE set)
            PROJECT_SRATE_USE : set to 1 if project samplerate is used
        """
        return RPR.GetSetProjectInfo(self.id, param_name, 0, False)

    def get_play_rate(self, position):
        """
        Return project play rate at a given position.

        Parameters
        ----------
        position : float
            Position in seconds.

        Returns
        -------
        play_rate : float
            Play rate at the given position.

        See also
        --------
        Project.play_rate
            Project play rate at the current position.
        """
        play_rate = RPR.Master_GetPlayRateAtTime(position, self.id)
        return play_rate

    def get_selected_item(self, index):
        """
        Return index-th selected item.

        Parameters
        ----------
        index : int
            Item index.

        Returns
        -------
        item : Item
            index-th selected item.
        """
        item_id = RPR.GetSelectedMediaItem(self.id, index)
        item = reapy.Item(item_id)
        return item

    def get_selected_track(self, index):
        """
        Return index-th selected track.

        Parameters
        ----------
        index : int
            Track index.

        Returns
        -------
        track : Track
            index-th selected track.
        """
        track_id = RPR.GetSelectedTrack(self.id, index)
        track = reapy.Track(track_id)
        return track

    def get_ext_state(self, section, key, pickled=False):
        """
        Return external state of project.

        Parameters
        ----------
        section : str
        key : str
        pickled: bool
            Whether data was pickled or not.

        Returns
        -------
        value : str
            If key or section does not exist an empty string is returned.
        """
        value = RPR.GetProjExtState(self.id, section, key, "", 2**31 - 1)[4]
        if value and pickled:
            value = pickle.loads(codecs.decode(value.encode(), "base64"))
        return value

    def glue_items(self, within_time_selection=False):
        """
        Glue items (action shortcut).

        Parameters
        ----------
        within_time_selection : bool
            If True, glue items within time selection.
        """
        action_id = 41588 if within_time_selection else 40362
        self.perform_action(action_id)

    @property
    def has_valid_id(self):
        """
        Whether ReaScript ID is still valid.

        For instance, if project has been closed, ID will not be valid
        anymore.

        :type: bool
        """
        return bool(RPR.ValidatePtr(*self._get_pointer_and_name()))

    @property
    def is_dirty(self):
        """
        Whether project is dirty (i.e. needing save).

        :type: bool
        """
        is_dirty = RPR.IsProjectDirty(self.id)
        return is_dirty

    @property
    def is_current_project(self):
        """
        Whether project is current project.

        :type: bool
        """
        is_current = self == Project()
        return is_current

    @property
    def is_paused(self):
        """
        Return whether project is paused.

        :type: bool
        """
        return bool(RPR.GetPlayStateEx(self.id) & 2)

    @property
    def is_playing(self):
        """
        Return whether project is playing.

        :type: bool
        """
        return bool(RPR.GetPlayStateEx(self.id) & 1)

    @property
    def is_recording(self):
        """
        Return whether project is recording.

        :type: bool
        """
        return bool(RPR.GetPlayStateEx(self.id) & 4)

    @reapy.inside_reaper()
    @property
    def is_stopped(self):
        """
        Return whether project is stopped.

        :type: bool
        """
        return not self.is_playing and not self.is_paused

    @reapy.inside_reaper()
    @property
    def items(self):
        """
        List of items in project.

        :type: list of Item
        """
        n_items = self.n_items
        item_ids = [RPR.GetMediaItem(self.id, i) for i in range(n_items)]
        return list(map(reapy.Item, item_ids))

    @property
    def length(self):
        """
        Project length in seconds.

        :type: float
        """
        length = RPR.GetProjectLength(self.id)
        return length

    @reapy.inside_reaper()
    @property
    def last_touched_fx(self):
        """
        Last touched FX and corresponding parameter index.

        :type: FX, int or NoneType, NoneType

        Notes
        -----
        Only Track FX are detected by this property. If last touched
        FX is a Take FX, this property is ``(None, None)``.

        Examples
        --------
        >>> fx, index = project.last_touched_fx
        >>> fx.name
        'VSTi: ReaSamplOmatic5000 (Cockos)'
        >>> fx.params[index].name
        "Volume"
        """
        if not self.is_current_project:
            fx, index = None, None
        else:
            res = RPR.GetLastTouchedFX(0, 0, 0)
            if not res[0]:
                fx, index = None, None
            else:
                if res[1]:
                    track = self.tracks[res[1] - 1]
                else:
                    track = self.master_track
                fx, index = track.fxs[res[2]], res[3]
        return fx, index

    def make_current_project(self):
        """
        Set project as current project.

        Can also be used as a context manager to temporarily set
        the project as current project and then go back to the original
        situation.

        Examples
        --------
        >>> p1 = reapy.Project()  # current project
        >>> p2 = reapy.Project(1)  # other project
        >>> p2.make_current_project()  # now p2 is current project
        >>> with p1.make_current_project():
        ...     do_something()  # current project is temporarily p1
        >>> # and p2 is current project again
        """
        return _MakeCurrentProject(self)

    def mark_dirty(self):
        """
        Mark project as dirty (i.e. needing save).
        """
        RPR.MarkProjectDirty(self.id)

    @reapy.inside_reaper()
    @property
    def markers(self):
        """
        List of project markers.

        :type: list of reapy.Marker
        """
        ids = [
            RPR.EnumProjectMarkers2(self.id, i, 0, 0, 0, 0, 0)
            for i in range(self.n_regions + self.n_markers)
        ]
        return [reapy.Marker(self, i[0]) for i in ids if not i[3]]

    @property
    def master_track(self):
        """
        Project master track.

        :type: Track
        """
        track_id = RPR.GetMasterTrack(self.id)
        master_track = reapy.Track(track_id)
        return master_track

    @reapy.inside_reaper()
    def mute_all_tracks(self, mute=True):
        """
        Mute or unmute all tracks.

        Parameters
        ----------
        mute : bool, optional
            Whether to mute or unmute all tracks (default=True).

        See also
        --------
        Project.unmute_all_tracks
        """
        with self.make_current_project():
            RPR.MuteAllTracks(mute)

    @property
    def n_items(self):
        """
        Number of items in project.

        :type: int
        """
        n_items = RPR.CountMediaItems(self.id)
        return n_items

    @property
    def n_markers(self):
        """
        Number of markers in project.

        :type: int
        """
        n_markers = RPR.CountProjectMarkers(self.id, 0, 0)[2]
        return n_markers

    @property
    def n_regions(self):
        """
        Number of regions in project.

        :type: int
        """
        n_regions = RPR.CountProjectMarkers(self.id, 0, 0)[3]
        return n_regions

    @property
    def n_selected_items(self):
        """
        Number of selected media items.

        :type: int
        """
        n_items = RPR.CountSelectedMediaItems(self.id)
        return n_items

    @property
    def n_selected_tracks(self):
        """
        Number of selected tracks in project (excluding master).

        :type: int
        """
        n_tracks = RPR.CountSelectedTracks2(self.id, False)
        return n_tracks

    @property
    def n_tempo_markers(self):
        """
        Number of tempo/time signature markers in project.

        :type: int
        """
        n_tempo_markers = RPR.CountTempoTimeSigMarkers(self.id)
        return n_tempo_markers

    @property
    def n_tracks(self):
        """
        Number of tracks in project.

        :type: int
        """
        n_tracks = RPR.CountTracks(self.id)
        return n_tracks

    @property
    def name(self):
        """
        Project name.

        :type: str
        """
        _, name, _ = RPR.GetProjectName(self.id, "", 2048)
        return name

    def open(self, in_new_tab=False):
        """
        Open project, if it was closed by Project.close.

        Parameters
        ----------
        in_new_tab : bool, optional
            whether should be opened in new tab

        Raises
        ------
        RuntimeError
            If hasn't been closed by Project.close yet
        """
        if self._filename is None:
            raise RuntimeError("project hasn't been closed")
        self.id = reapy.open_project(self._filename, in_new_tab).id

    def pause(self):
        """
        Hit pause button.
        """
        RPR.OnPauseButtonEx(self.id)

    @property
    def path(self):
        """
        Project path.

        :type: str
        """
        _, path, _ = RPR.GetProjectPathEx(self.id, "", 2048)
        return path

    def perform_action(self, action_id):
        """
        Perform action with ID `action_id` in the main Actions section.

        Parameters
        ----------
        action_id : int
            Action ID in the main Actions section.
        """
        RPR.Main_OnCommandEx(action_id, 0, self.id)

    def play(self):
        """
        Hit play button.
        """
        RPR.OnPlayButtonEx(self.id)

    @property
    def play_position(self):
        """
        Latency-compensated actual-what-you-hear position in seconds.

        :type: float

        See also
        --------
        Project.buffer_position
            Position of next audio block being processed.
        """
        return RPR.GetPlayPositionEx(self.id)

    @property
    def play_rate(self):
        """
        Project play rate at the cursor position.

        :type: float

        See also
        --------
        Project.get_play_rate
            Return project play rate at a specified time.
        """
        play_rate = RPR.Master_GetPlayRate(self.id)
        return play_rate

    @reapy.inside_reaper()
    def record(self):
        """Hit record button."""
        with self.make_current_project():
            reapy.perform_action(1013)

    def redo(self):
        """
        Redo last action.

        Raises
        ------
        RedoError
            If impossible to redo.
        """
        success = RPR.Undo_DoRedo2(self.id)
        if not success:
            raise RedoError

    @reapy.inside_reaper()
    @property
    def regions(self):
        """
        List of project regions.

        :type: list of reapy.Region
        """
        ids = [
            RPR.EnumProjectMarkers2(self.id, i, 0, 0, 0, 0, 0)
            for i in range(self.n_regions + self.n_markers)
        ]
        return [reapy.Region(self, i[0]) for i in ids if i[3]]

    def save(self, force_save_as=False):
        """
        Save project.

        Parameters
        ----------
        force_save_as : bool
            Force using "Save as" instead of "Save".
        """
        RPR.Main_SaveProject(self.id, force_save_as)

    def select(self, start, end=None, length=None):
        if end is None:
            message = "Either `end` or `length` must be specified."
            assert length is not None, message
            end = start + length
        self.time_selection = start, end

    def select_all_items(self, selected=True):
        """
        Select or unselect all items, depending on `selected`.

        Parameters
        ----------
        selected : bool
            Whether to select or unselect items.
        """
        RPR.SelectAllMediaItems(self.id, selected)

    def select_all_tracks(self):
        """Select all tracks."""
        self.perform_action(40296)

    @property
    def selected_envelope(self):
        """
        Project selected envelope.

        :type: reapy.Envelope or None
        """
        envelope_id = RPR.GetSelectedTrackEnvelope(self.id)
        envelope = None if envelope_id == 0 else reapy.Envelope(envelope_id)
        return envelope

    @reapy.inside_reaper()
    @property
    def selected_items(self):
        """
        List of all selected items.

        :type: list of Item

        See also
        --------
        Project.get_selected_item
            Return a specific selected item.
        """
        return [
            reapy.Item(RPR.GetSelectedMediaItem(self.id, i))
            for i in range(self.n_selected_items)
        ]

    @reapy.inside_reaper()
    @property
    def selected_tracks(self):
        """
        List of selected tracks (excluding master).

        :type: list of Track
        """
        return [
            reapy.Track(RPR.GetSelectedTrack(self.id, i))
            for i in range(self.n_selected_tracks)
        ]

    @selected_tracks.setter
    def selected_tracks(self, tracks):
        self.unselect_all_tracks()
        for track in tracks:
            track.select()

    def set_info_string(self, param_name, param_string):
        """
        Parameters
        ----------
        param_name : str
            MARKER_GUID:X : get the GUID (unique ID) of the marker or region
                with index X, where X is the index passed to
                EnumProjectMarkers, not necessarily the displayed number
            RECORD_PATH :
                recording directory -- may be blank or a relative path,
                to get the effective path see GetProjectPathEx()
            RENDER_FILE : render directory
            RENDER_PATTERN : render file name (may contain wildcards)
            RENDER_FORMAT : base64-encoded sink configuration
                (see project files, etc). Callers can also pass a simple
                4-byte string (non-base64-encoded), e.g. "evaw" or "l3pm",
                to use default settings for that sink type.
            RENDER_FORMAT2 : base64-encoded secondary sink configuration.
                Callers can also pass a simple 4-byte string (non-base64-encoded),
                e.g. "evaw" or "l3pm", to use default settings for
                that sink type, or "" to disable secondary render.
                Formats available on this machine:
                "wave" "aiff" "iso " "ddp " "flac" "mp3l" "oggv" "OggS"
                "FFMP" "GIF " "LCF " "wvpk"
        param_string : str
        """
        RPR.GetSetProjectInfo_String(
            self.id, param_name, param_string, True
        )

    def set_info_value(self, param_name, param_value):
        """
        Parameters
        ----------
        param_name : str
            RENDER_SETTINGS : &(1|2)=0:master mix, &1=stems+master mix,
                &2=stems only, &4=multichannel tracks to multichannel files,
                &8=use render matrix, &16=tracks with only mono media
                to mono files, &32=selected media items,
                &64=selected media items via master
            RENDER_BOUNDSFLAG : 0=custom time bounds, 1=entire project,
                2=time selection, 3=all project regions,
                4=selected media items, 5=selected project regions
            RENDER_CHANNELS : number of channels in rendered file
            RENDER_SRATE : sample rate of rendered file
                (or 0 for project sample rate)
            RENDER_STARTPOS : render start time when RENDER_BOUNDSFLAG=0
            RENDER_ENDPOS : render end time when RENDER_BOUNDSFLAG=0
            RENDER_TAILFLAG : apply render tail setting when rendering:
                &1=custom time bounds, &2=entire project, &4=time selection,
                &8=all project regions, &16=selected media items,
                &32=selected project regions
            RENDER_TAILMS : tail length in ms to render
                (only used if RENDER_BOUNDSFLAG and RENDER_TAILFLAG are set)
            RENDER_ADDTOPROJ : 1=add rendered files to project
            RENDER_DITHER : &1=dither, &2=noise shaping, &4=dither stems,
                &8=noise shaping on stems
            PROJECT_SRATE : samplerate (ignored unless PROJECT_SRATE_USE set)
            PROJECT_SRATE_USE : set to 1 if project samplerate is used
        param_value : float
        """
        RPR.GetSetProjectInfo(self.id, param_name, param_value, True)

    def set_ext_state(self, section, key, value, pickled=False):
        """
        Set external state of project.

        Parameters
        ----------
        section : str
        key : str
        value : Union[Any, str]
            State value. Will be dumped to str using either `pickle` if
            `pickled` is `True` or `json`. Length of the dumped value
            must not be over 2**31 - 2.
        pickled : bool, optional
            Data will be pickled with the last version if True.
            If you using mypy as type checker, typing_extensions.Literal[True]
            has to be used for `pickled`.

        Raises
        ------
        ValueError
            If dumped `value` has length over 2**31 - 2.
        """
        if pickled:
            value = pickle.dumps(value)
            value = codecs.encode(value, "base64").decode()
        if len(value) > 2**31 - 2:
            message = (
                "Dumped value length is {:,d}. It must not be over "
                "2**31 - 2."
            )
            raise ValueError(message.format(len(value)))
        RPR.SetProjExtState(self.id, section, key, value)

    @reapy.inside_reaper()
    def solo_all_tracks(self):
        """
        Solo all tracks in project.

        See also
        --------
        Project.unsolo_all_tracks
        """
        with self.make_current_project():
            RPR.SoloAllTracks(1)

    def stop(self):
        """
        Hit stop button.
        """
        RPR.OnStopButtonEx(self.id)

    @reapy.inside_reaper()
    @property
    def time_selection(self):
        """
        Project time selection.

        time_selection : reapy.TimeSelection

        Can be set and deleted as follows:

        >>> project = reapy.Project()
        >>> project.time_selection = 3, 8  # seconds
        >>> del project.time_selection
        """
        time_selection = reapy.TimeSelection(self)
        return time_selection

    @time_selection.setter
    def time_selection(self, selection):
        """
        Set time selection bounds.

        Parameters
        ----------
        selection : (float, float)
            Start and end of new time selection in seconds.
        """
        self.time_selection._set_start_end(*selection)

    @time_selection.deleter
    def time_selection(self):
        """
        Delete current time selection.
        """
        self.time_selection._set_start_end(0, 0)

    @property
    def time_signature(self):
        """
        Project time signature.

        This does not reflect tempo envelopes but is purely what is set in the
        project settings.

        bpm : float
            Project BPM (beats per minute)
        bpi : float
            Project BPI (numerator of time signature)
        """
        _, bpm, bpi = RPR.GetProjectTimeSignature2(self.id, 0, 0)
        return bpm, bpi

    def time_to_beats(self, time):
        """
        Convert time in seconds to beats.

        Parameters
        ----------
        time : float
            Time in seconds.

        Returns
        -------
        beats : float
            Time in beats.

        See also
        --------
        Projecr.beats_to_time
        """
        beats = RPR.TimeMap2_timeToQN(self.id, time)
        return beats

    @property
    def tracks(self):
        """
        List of project tracks.

        :type: TrackList
        """
        return reapy.TrackList(self)

    def undo(self):
        """
        Undo last action.

        Raises
        ------
        UndoError
            If impossible to undo.
        """
        success = RPR.Undo_DoUndo2(self.id)
        if not success:
            raise UndoError

    def unmute_all_tracks(self):
        """
        Unmute all tracks.
        """
        self.mute_all_tracks(mute=False)

    def unselect_all_tracks(self):
        """Unselect all tracks."""
        self.perform_action(40297)

    @reapy.inside_reaper()
    def unsolo_all_tracks(self):
        """
        Unsolo all tracks in project.

        See also
        --------
        Project.solo_all_tracks
        """
        with self.make_current_project():
            RPR.SoloAllTracks(0)


class _MakeCurrentProject:

    """Context manager used by Project.make_current_project."""

    def __init__(self, project):
        self.current_project = self._make_current_project(project)

    @staticmethod
    @reapy.inside_reaper()
    def _make_current_project(project):
        """Set current project and return the previous current project."""
        current_project = reapy.Project()
        RPR.SelectProjectInstance(project.id)
        return current_project

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Test for valid ID in case project has been closed since __enter__
        if self.current_project.has_valid_id:
            self.current_project.make_current_project()



"""Define reapy.defer and reapy.at_exit."""

import reapy

import os
import sys
import tempfile


class Deferrer:

    """Class to register and run deferred calls."""

    _next_call_id = 0
    _callbacks = {}
    _args = {}
    _kwargs = {}

    def _get_new_call_id(self):
        new_id = self._next_call_id
        Deferrer._next_call_id += 1
        return new_id

    def _wrapped_open(self, *args, **kwargs):
        if args[0] == os.path.join(tempfile.gettempdir(), "reascripterr.txt"):
            return ReaperConsole()
        return open(*args, **kwargs)

    def defer(self, callback, args, kwargs, at_exit=False):
        call_id = self._get_new_call_id()
        Deferrer._callbacks[call_id] = callback
        Deferrer._args[call_id] = args
        Deferrer._kwargs[call_id] = kwargs
        code = "import sys; sys.modules[{}].Deferrer().run({})".format(
            repr(__name__), call_id
        )
        sys.modules["__main__"].open = self._wrapped_open
        if at_exit:
            sys.modules["__main__"].RPR_atexit(code)
        else:
            sys.modules["__main__"].RPR_defer(code)

    def run(self, call_id):
        # Get callback and run it
        callback = self._callbacks[call_id]
        args = self._args[call_id]
        kwargs = self._kwargs[call_id]
        callback(*args, **kwargs)
        # Clean up
        sys.modules["__main__"].open = open
        for x in Deferrer._callbacks, Deferrer._args, Deferrer._kwargs:
            del x[call_id]


class ReaperConsole:

    """File-like wrapper around the Reaper Console."""

    def close(self):
        pass

    def flush(self):
        pass

    def write(self, *args, **kwargs):
        reapy.print(*args, **kwargs)


def at_exit(f, *args, **kwargs):
    """
    Make REAPER call a function after script execution.

    The function is also called if excution is terminated by user.

    Parameters
    ----------
    f : callable
        Function to be called later.
    args : tuple, optional
        Positional arguments to pass to ``f``.
    kwargs : dict, optional
        Keyword arguments to pass to ``f``.

    Raises
    ------
    AssertionError
        When called from outside REAPER.

    Examples
    --------
    Typical use case of ``at_exit`` is cleaning up after a ``defer``
    loop.

    The following example opens a file and starts a loop that
    indefinitely writes integers to that file. Since we want the file
    to be closed when the user terminates script execution, call to
    its ``close`` method is deferred to ``reapy.at_exit``.

    >>> import reapy
    >>> file = open("somefile.txt", "w")
    >>> def stupid_loop(i):
    ...     file.write(i)
    ...     reapy.defer(stupid_loop, i + 1)
    ...
    >>> reapy.at_exit(file.close)
    >>> stupid_loop(0)
    """
    message = "reapy.at_exit can only be called inside REAPER."
    assert reapy.is_inside_reaper(), message
    Deferrer().defer(f, args, kwargs, at_exit=True)


def defer(f, *args, **kwargs):
    """
    Make REAPER call a function later.

    Parameters
    ----------
    f : callable
        Function to be called later.
    args : tuple, optional
        Positional arguments to pass to ``f``.
    kwargs : dict, optional
        Keyword arguments to pass to ``f``.

    Raises
    ------
    AssertionError
        When called from outside REAPER.

    Notes
    -----
    The average time before a defered call is actually run is about
    0.03 seconds (around 30 defered calls are allowed per second).

    Examples
    --------
    Typical use case of ``defer`` is running loops that don't block
    REAPER GUI.

    The following example creates a loop that indefinitely prints
    integers to the REAPER console, without blocking REAPER GUI.

    >>> import reapy
    >>> def stupid_loop(i):
    ...     reapy.print(i)
    ...     reapy.defer(stupid_loop, i + 1)
    ...
    >>> stupid_loop(0)
    """
    # Check we are inside REAPER
    message = "reapy.defer can only be called inside REAPER."
    assert reapy.is_inside_reaper(), message
    Deferrer().defer(f, args, kwargs)



import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject
import typing as ty


class Item(ReapyObject):

    _class_name = "Item"
    id: int

    def __init__(self, id: int) -> None:
        ...

    def __eq__(self, other: object) -> bool:
        ...

    @property
    def _args(self) -> ty.Tuple[int]:
        ...

    @property
    def active_take(self) -> reapy.Take:
        """
        Return the active take of the item.

        Returns
        -------
        take : Take
            Active take of the item.
        """
        ...

    def add_take(self) -> reapy.Take:
        """
        Create and return a new take in item.

        Returns
        -------
        take : Take
            New take in item.
        """
        ...

    def delete(self) -> None:
        """Delete item."""
        ...

    def get_info_value(self, param_name: str) -> float:
        ...

    def get_take(self, index: int) -> reapy.Take:
        """
        Return index-th take of item.

        Parameters
        ----------
        index : int
            Take index.

        Returns
        -------
        take : Take
            index-th take of media item.
        """
        ...

    @property
    def has_valid_id(self) -> bool:
        """
        Whether ReaScript ID is still valid.

        For instance, if item has been deleted, ID will not be valid
        anymore.

        :type: bool
        """

    @property
    def is_selected(self) -> bool:
        """
        Return whether item is selected.

        Returns
        -------
        is_selected : bool
            Whether item is selected.
        """
        ...

    @property
    def length(self) -> float:
        """
        Return item length in seconds.

        Returns
        -------
        length : float
            Item length in seconds.
        """
        ...

    @length.setter
    def length(self, length: float) -> None:
        """
        Set item length.

        Parameters
        ----------
        length : float
            New item length in seconds.
        """
        ...

    @property
    def n_takes(self) -> int:
        """
        Return the number of takes of media item.

        Returns
        -------
        n_takes : int
            Number of takes of media item.
        """
        ...

    @property
    def position(self) -> float:
        """
        Return item position in seconds.

        Returns
        -------
        position : float
            Item position in seconds.
        """
        ...

    @position.setter
    def position(self, position: float) -> None:
        """
        Set media item position to `position`.

        Parameters
        ----------
        position : float
            New item position in seconds.
        """
        ...

    @property
    def project(self) -> reapy.Project:
        """
        Return item parent project.

        Returns
        -------
        project : Project
            Item parent project.
        """
        ...

    def set_info_value(self, param_name: str, value: float) -> bool:
        ...

    def split(self, position: float) -> ty.Tuple[Item, Item]:
        """
        Split item and return left and right parts.

        Parameters
        ----------
        position : float
            Split position in seconds.

        Returns
        -------
        left, right : Item
            Left and right parts of the split.
        """
        ...

    @property
    def takes(self) -> ty.List[reapy.Take]:
        """
        Return list of all takes of media item.

        Returns
        -------
        takes : list of Take
            List of all takes of media item.
        """
        ...

    @property
    def track(self) -> reapy.Track:
        """
        Parent track of item.

        Set it by passing a track, or a track index.

        :type: Track

        Examples
        --------
        >>> track0, track1 = project.tracks[0:2]
        >>> item = track0.items[0]
        >>> item.track == track0
        True
        >>> item.track = track1  # Move to track 1
        >>> item.track = 0  # Move to track 0
        """
        ...

    @track.setter
    def track(self, track: ty.Union[int, reapy.Track]) -> None:
        ...

    def update(self) -> None:
        """Update item in REAPER interface."""
        ...



import reapy
import reapy.reascript_api as RPR


def get_active_editor():
    """
    Return active MIDI editor, or None if no editor is active.

    Returns
    -------
    editor : MIDIEditor or None
        Active MIDI editor, or None if no editor is active.
    """
    editor = reapy.MIDIEditor(RPR.MIDIEditor_GetActive())
    if not editor._is_defined:
        editor = None
    return editor


@reapy.inside_reaper()
def get_input_names():
    """
    Return names of all input channels.

    Returns
    -------
    names : list of str
        Names of input channels.
    """
    n_channels = reapy.midi.get_n_inputs()
    return [RPR.GetMIDIInputName(i, "", 2048)[2] for i in range(n_channels)]


def get_max_inputs():
    """
    Return maximum number of MIDI inputs.

    Returns
    -------
    max_inputs : int
        Maximum number of MIDI inputs.
    """
    max_inputs = RPR.GetMaxMidiInputs()
    return max_inputs


def get_max_outputs():
    """
    Return maximum number of MIDI outputs.

    Returns
    -------
    max_outputs : int
        Maximum number of MIDI outputs.
    """
    max_outputs = RPR.GetMaxMidiOutputs()
    return max_outputs


def get_n_inputs():
    """
    Return number of MIDI inputs.

    Returns
    -------
    n_inputs : int
        Number of MIDI inputs.
    """
    n_inputs = RPR.GetNumMIDIInputs()
    return n_inputs


def get_n_outputs():
    """
    Return number of MIDI outputs.

    Returns
    -------
    n_outputs : int
        Number of MIDI outputs.
    """
    n_outputs = RPR.GetNumMIDIOutputs()
    return n_outputs


@reapy.inside_reaper()
def get_output_names():
    """
    Return names of all output channels.

    Returns
    -------
    names : list of str
        Names of output channels.
    """
    n_channels = reapy.midi.get_n_outputs()
    return [RPR.GetMIDIOutputName(i, "", 2048)[2] for i in range(n_channels)]


def reinit():
    """Reset all MIDI devices."""
    RPR.midi_reinit()



"""Audio handling functions."""

import reapy
import reapy.reascript_api as RPR
import typing as ty


def get_input_latency(unit: str = "second") -> float:
    """
    Return input latency.

    Parameters
    ----------
    unit : {"sample", "second"}
        Whether to return latency in samples or seconds
        (default="second").

    Returns
    -------
    latency : float
        Input latency.
    """
    ...


@reapy.inside_reaper()
def get_input_names() -> ty.List[str]:
    """
    Return names of all input channels.

    Returns
    -------
    names : list of str
        Names of input channels.
    """
    ...


def get_n_inputs() -> int:
    """
    Return number of audio inputs.

    Returns
    -------
    n_inputs : int
        Number of audio inputs.
    """
    ...


def get_n_outputs() -> int:
    """
    Return number of audio outputs.

    Returns
    -------
    n_outputs : int
        Number of audio outputs.
    """
    ...


def get_output_latency(unit: str = "second") -> float:
    """
    Return output latency.

    Parameters
    ----------
    unit : {"sample", "second"}
        Whether to return latency in samples or seconds
        (default="second").

    Returns
    -------
    latency : float
        Output latency.
    """
    ...


@reapy.inside_reaper()
def get_output_names() -> ty.Tuple[str, ...]:
    """
    Return names of all output channels.

    Returns
    -------
    names : list of str
        Names of output channels.
    """
    ...


def init() -> None:
    """
    Open all audio and MIDI devices (if not opened).
    """
    ...


def is_prebuffer() -> bool:
    """
    Return whether audio is in pre-buffer (threadsafe).

    Returns
    -------
    is_prebuffer : bool
        Whether audio is in pre-buffer.
    """
    ...


def is_running() -> bool:
    """
    Return whether audio is running (threadsafe).

    Returns
    -------
    is_running : bool
        Whether audio is running.
    """
    ...


def quit() -> None:
    """
    Close all audio and MIDI devices (if opened).
    """
    ...



import typing as ty


def map(
    function: ty.Callable,
    *iterables: ty.Iterable,
    constants: ty.Mapping[str, ty.Any] = ...,
    kwargs_iterable: ty.Optional[ty.Iterable[ty.Mapping[str, ty.Any]]] = ...
) -> list:
    ...



import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject


class Region(ReapyObject):

    _class_name = "Region"

    def __init__(
        self, parent_project=None, index=None, parent_project_id=None
    ):
        if parent_project_id is None:
            message = (
                "One of `parent_project` or `parent_project_id` must be "
                "specified."
            )
            assert parent_project is not None, message
            parent_project_id = parent_project.id
        self.project_id = parent_project_id
        self.index = index

    @reapy.inside_reaper()
    def _get_enum_index(self):
        """
        Return region index as needed by RPR.EnumProjectMarkers2.
        """
        return next(
            i for i, r in enumerate(reapy.Project(self.project_id).regions)
            if r.index == self.index
        )

    @property
    def _kwargs(self):
        return {
            "index": self.index, "parent_project_id": self.project_id
        }

    def add_rendered_track(self, track):
        """
        Add track to region render matrix for this region.

        Parameters
        ----------
        track : Track
            Track to add.

        See also
        --------
        Region.add_rendered_tracks
            Efficiently add several tracks to region render matrix.
        Region.remove_rendered_track
        Region.remove_rendered_tracks
        """
        RPR.SetRegionRenderMatrix(self.project_id, self.index, track.id, 1)

    @reapy.inside_reaper()
    def add_rendered_tracks(self, tracks):
        """
        Efficiently add  several tracks to region render matrix.

        Parameters
        ----------
        tracks : list of Track
            Tracks to add.

        See also
        --------
        Region.remove_rendered_tracks
        """
        for track in tracks:
            self.add_rendered_track(track)

    @reapy.inside_reaper()
    @property
    def end(self):
        """
        Region end.

        :type: float
            Region end in seconds.
        """
        index = self._get_enum_index()
        args = self.project_id, index, 0, 0, 0, 0, 0
        return RPR.EnumProjectMarkers2(*args)[5]

    @end.setter
    def end(self, end):
        """
        Set region end.

        Parameters
        ----------
        end : float
            region end in seconds.
        """
        args = self.project_id, self.index, True, self.start, end, ""
        RPR.SetProjectMarker2(*args)

    def delete(self):
        """
        Delete region.
        """
        RPR.DeleteProjectMarker(self.project_id, self.index, True)

    def remove_rendered_track(self, track):
        """
        Remove track from region render matrix for this region.

        Parameters
        ----------
        track : Track
            Track to remove.

        See also
        --------
        Region.add_rendered_tracks
        Region.remove_rendered_track
        Region.remove_rendered_tracks
            Efficiently remove several tracks from render matrix.
        """
        RPR.SetRegionRenderMatrix(self.project_id, self.index, track.id, -1)

    @reapy.inside_reaper()
    def remove_rendered_tracks(self, tracks):
        """
        Efficiently remove  several tracks from region render matrix.

        Parameters
        ----------
        tracks : list of Track
            Tracks to remove.

        See also
        --------
        Region.add_rendered_tracks
        """
        for track in tracks:
            self.remove_rendered_track(track)

    @reapy.inside_reaper()
    @property
    def rendered_tracks(self):
        """
        List of tracks for this region in region render matrix.

        :type: list of Track
        """
        i = 0
        tracks = []
        while i == 0 or tracks[-1]._is_defined:
            track_id = RPR.EnumRegionRenderMatrix(
                self.project_id, self.index, i
            )
            tracks.append(reapy.Track(track_id))
            i += 1
        return tracks[:-1]

    @reapy.inside_reaper()
    @property
    def start(self):
        """
        Region start.

        :type: float
        """
        args = self.project_id, self._get_enum_index(), 0, 0, 0, 0, 0
        return RPR.EnumProjectMarkers2(*args)[4]

    @start.setter
    def start(self, start):
        """
        Set region start.

        Parameters
        ----------
        start : float
            region start in seconds.
        """
        RPR.SetProjectMarker2(
            self.project_id, self.index, 1, start, self.end, ""
        )



from .marker import Marker
from .project import Project
from .region import Region
from .time_selection import TimeSelection
__all__ = [
    'Marker',
    'Project',
    'Region',
    'TimeSelection',
]



import importlib
import sys
import warnings
import typing as ty
from builtins import BaseException
from types import TracebackType

import reapy
import reapy.config
from reapy import errors
from . import client, web_interface


CLIENT: ty.Optional[client.Client]
CLIENTS: ty.Dict[ty.Optional[str], ty.Optional[client.Client]]


def get_selected_client() -> ty.Optional[client.Client]:
    ...


def get_selected_machine_host() -> ty.Optional[str]:
    """Return host of the currently selected machine.

    Returns
    -------
    host : str or None
        None is returned when running from inside REAPER and
        no slave machine is selected.
    """
    ...


def reconnect() -> None:
    """
    Reconnect to REAPER ReaScript API.

    Examples
    --------
    Assume no REAPER instance is active.
    >>> import reapy
    errors.DisabledDistAPIWarning: Can't reach distant API. Please start REAPER, or
    call reapy.config.enable_dist_api() from inside REAPER to enable distant
    API.
      warnings.warn(errors.DisabledDistAPIWarning())
    >>> p = reapy.Project()  # Results in error
    Traceback (most recent call last):
      File "<string>", line 1, in <module>
      File "reapy\\core\\project\\project.py", line 26, in __init__
        id = RPR.EnumProjects(index, None, 0)[0]
    AttributeError: module 'reapy.reascript_api' has no attribute 'EnumProjects'
    >>> # Now start REAPER
    ...
    >>> reapy.reconnect()
    >>> p = reapy.Project()  # No error!
    """
    ...


class connect:

    """Connect to slave machine.

    reapy instructions will now be run on the selected machine.
    If used as a context manager, the slave machine will only be
    selected in the corresponding context.

    Parameters
    ----------
    host : str, optional
        Slave machine host. If None, selects default ``reapy``
        behavior (i.e. local REAPER instance).

    See also
    --------
    ``connect_to_default_machine``
        Connect to default slave machine (i.e. local REAPER instance).
    """

    previous_client: ty.Optional[client.Client]

    def __init__(self, host: ty.Optional[str] = None) -> None:
        ...

    def __enter__(self) -> 'connect':
        ...

    def __exit__(self,
                 type: ty.Optional[ty.Type[BaseException]],
                 value: ty.Optional[BaseException],
                 traceback: ty.Optional[TracebackType]) -> None:
        ...


class connect_to_default_machine(connect):

    """Select default slave machine (i.e. local REAPER instance)."""

    def __init__(self) -> None:
        ...


def register_machine(host: str) -> None:
    """Register a slave machine.

    Parameters
    ----------
    host : str
        Slave machine host (e.g. ``"localhost"``).

    See also
    --------
    ``reapy.connect``
    """
    ...



"""
Enable ```reapy`` distant API.

Running this ReaScript from inside REAPER allows to import ``reapy``
from outside. It creates a persistent Web Interface inside REAPER and
adds the ReaScript ``reapy.reascripts.activate_reapy_server`` to the
Actions list. Importing ``reapy`` from outside REAPER will trigger
the latter **via** the Web Interface.
"""

if __name__ == "__main__":
    import reapy
    reapy.config.enable_dist_api()



import contextlib
import functools
import typing as ty

import reapy
import reapy.config
from reapy.errors import DisabledDistAPIError, DisabledDistAPIWarning
if not reapy.is_inside_reaper():
    try:
        from .network import Client, WebInterface
        _WEB_INTERFACE = WebInterface(reapy.config.WEB_INTERFACE_PORT)
        _CLIENT = Client(_WEB_INTERFACE.get_reapy_server_port())
    except DisabledDistAPIError:
        import warnings
        warnings.warn(DisabledDistAPIWarning())
        _CLIENT = None  # type:ignore

FuncType = ty.Callable[..., ty.Any]
F = ty.TypeVar('F', bound=FuncType)


def dist_api_is_enabled() -> bool:
    """Return whether reapy can reach REAPER from the outside."""
    ...


class inside_reaper(contextlib.ContextDecorator):
    """
    Context manager for efficient calls from outside REAPER.

    It can also be used as a function decorator.

    Examples
    --------
    Instead of running:

    >>> project = reapy.Project()
    >>> l = [project.bpm for i in range(1000)

    which takes around 30 seconds, run:

    >>> project = reapy.Project()
    >>> with reapy.inside_reaper():
    ...     l = [project.bpm for i in range(1000)
    ...

    which takes 0.1 seconds!

    Example usage as decorator:

    >>> @reapy.inside_reaper()
    ... def add_n_tracks(n):
    ...     for x in range(n):
    ...         reapy.Project().add_track()

    """

    def __call__(self,  # type:ignore
                 func: F,
                 encoded_func: ty.Optional[ty.Dict[str, object]] = None
                 ) -> F:
        ...

    def __enter__(self) -> None:
        ...

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:  # type: ignore
        ...


class DistProperty(property):

    _inside_reaper: inside_reaper

    @classmethod
    def from_property(cls, p: property) -> property:
        ...

    @staticmethod
    def _encode(f: ty.Callable[..., ty.Any],
                method_name: str) -> ty.Dict[str, object]:
        ...

    def getter(self, fget: F) -> property:
        ...

    def setter(self, fset: F) -> property:
        ...

    def deleter(self, fdel: F) -> property:
        ...


def reconnect() -> None:
    """
    Reconnect to REAPER ReaScript API.

    This function has no effect from inside REAPER.

    Examples
    --------
    Assume no REAPER instance is active.
    >>> import reapy
    DisabledDistAPIWarning: Can't reach distant API. Please start REAPER, or
    call reapy.config.enable_dist_api() from inside REAPER to enable distant
    API.
      warnings.warn(DisabledDistAPIWarning())
    >>> p = reapy.Project()  # Results in error
    Traceback (most recent call last):
      File "<string>", line 1, in <module>
      File "C:\\Users\\despres\\Desktop\\reaper\\scripts\\reapy\\reapy\\core\\project\\project.py", line 26, in __init__
        id = RPR.EnumProjects(index, None, 0)[0]
    AttributeError: module 'reapy.reascript_api' has no attribute 'EnumProjects'
    >>> # Now start REAPER
    ...
    >>> reapy.reconnect()
    >>> p = reapy.Project()  # No error!
    """
    ...



from .item import Item
from .midi_event import CC, CCList, Note, NoteList
from .source import Source
from .take import Take
__all__ = [
    'Item',
    'CC',
    'CCList',
    'Note',
    'NoteList',
    'Source',
    'Take',
]



import reapy
import typing as ty
from typing_extensions import TypedDict

ReapyObjectDict = TypedDict(
    'ReapyObjectDict', {
        "__reapy__": bool,
        "class": str,
        "args": ty.Tuple[ty.Any, ...],
        "kwargs": ty.Dict[str, ty.Any]
    }
)


class ReapyObject:
    """Base class for reapy objects."""

    def __eq__(self, other: object) -> bool:
        ...

    def __repr__(self) -> str:
        ...

    @property
    def _args(self) -> ty.Tuple[ty.Any, ...]:
        ...

    def _get_pointer_and_name(self) -> ty.Tuple[int, str]:
        ...

    @property
    def _is_defined(self) -> bool:
        ...

    @property
    def _kwargs(self) -> ty.Dict[str, ty.Any]:
        ...

    def _to_dict(self) -> ReapyObjectDict:
        ...

class ReapyObjectList(ReapyObject):
    """Abstract class for list of ReapyObjects."""

    pass



import importlib
import sys
import warnings

import reapy
import reapy.config
from reapy import errors
from . import client, web_interface


CLIENT = None
CLIENTS = {None: None}


def get_selected_client():
    return CLIENT


def get_selected_machine_host():
    """Return host of the currently selected machine.

    Returns
    -------
    host : str or None
        None is returned when running from inside REAPER and
        no slave machine is selected.
    """
    return None if CLIENT is None else CLIENT.host


def reconnect():
    """
    Reconnect to REAPER ReaScript API.

    Examples
    --------
    Assume no REAPER instance is active.
    >>> import reapy
    errors.DisabledDistAPIWarning: Can't reach distant API. Please start REAPER, or
    call reapy.config.enable_dist_api() from inside REAPER to enable distant
    API.
      warnings.warn(errors.DisabledDistAPIWarning())
    >>> p = reapy.Project()  # Results in error
    Traceback (most recent call last):
      File "<string>", line 1, in <module>
      File "reapy\\core\\project\\project.py", line 26, in __init__
        id = RPR.EnumProjects(index, None, 0)[0]
    AttributeError: module 'reapy.reascript_api' has no attribute 'EnumProjects'
    >>> # Now start REAPER
    ...
    >>> reapy.reconnect()
    >>> p = reapy.Project()  # No error!
    """
    if not reapy.is_inside_reaper():
        host = get_selected_machine_host()
        if host is None:
            # We are outside REAPER, so this means initial import failed to
            # connect and we want to retry with default host (i.e. localhost)
            host = "localhost"
        try:
            del CLIENTS[host]
        except KeyError:
            pass
        connect(host)


class connect:

    """Connect to slave machine.

    reapy instructions will now be run on the selected machine.
    If used as a context manager, the slave machine will only be
    selected in the corresponding context.

    Parameters
    ----------
    host : str, optional
        Slave machine host. If None, selects default ``reapy``
        behavior (i.e. local REAPER instance).

    See also
    --------
    ``connect_to_default_machine``
        Connect to default slave machine (i.e. local REAPER instance).
    """

    def __init__(self, host=None):
        global CLIENT
        self.previous_client = CLIENT
        try:
            if host not in CLIENTS:
                register_machine(host)
            CLIENT = CLIENTS[host]
            if hasattr(reapy, 'reascript_api'):  # False during initial import
                importlib.reload(reapy.reascript_api)
        except errors.DisabledDistAPIError as e:
            if host and host != 'localhost':
                raise e
            warnings.warn(errors.DisabledDistAPIWarning())

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        global CLIENT
        CLIENT = self.previous_client
        importlib.reload(reapy.reascript_api)


class connect_to_default_machine(connect):

    """Select default slave machine (i.e. local REAPER instance)."""

    def __init__(self):
        super().__init__()


def register_machine(host):
    """Register a slave machine.

    Parameters
    ----------
    host : str
        Slave machine host (e.g. ``"localhost"``).

    See also
    --------
    ``reapy.connect``
    """
    if reapy.is_inside_reaper() and host == "localhost":
        msg = "A REAPER instance can not connect to istelf."
        raise errors.InsideREAPERError(msg)
    interface_port = reapy.config.WEB_INTERFACE_PORT
    interface = web_interface.WebInterface(interface_port, host)
    CLIENTS[host] = client.Client(interface.get_reapy_server_port(), host)


if not reapy.is_inside_reaper():
    connect("localhost")
    CLIENTS[None] = CLIENT



import reapy
import reapy.reascript_api as RPR
from reapy.core import ReapyObject, ReapyObjectList
from reapy.errors import DistError


class FXParam(float):

    """FX parameter."""

    def __init__(self, value, parent_list, index, functions):
        float.__init__(value)
        self.parent_list = parent_list
        self.index = index
        self.functions = functions

    def __new__(self, value, *args, **kwargs):
        return float.__new__(self, value)

    def add_envelope(self):
        """
        Create envelope for the parameter and return it.

        Returns
        -------
        envelope : Envelope
            New envelope for the parameter.

        Notes
        -----
        If the envelope already exists, the function returns it.
        """
        parent_fx = self.parent_list.parent_fx
        parent = parent_fx.parent
        if isinstance(parent, reapy.Track):
            callback = RPR.GetFXEnvelope
        else:  # Then it is a Take
            callback = self.functions["GetEnvelope"]
        envelope = reapy.Envelope(parent, callback(
            parent.id, parent_fx.index, self.index, True
        ))
        return envelope

    @property
    def envelope(self):
        """
        Parameter envelope (or None if it doesn't exist).

        :type: Envelope or NoneType
        """
        parent_fx = self.parent_list.parent_fx
        parent = parent_fx.parent
        if isinstance(parent, reapy.Track):
            callback = RPR.GetFXEnvelope
        else:  # Then it is a Take
            callback = self.functions["GetEnvelope"]
        envelope = reapy.Envelope(parent, callback(
            parent.id, parent_fx.index, self.index, False
        ))
        if not envelope._is_defined:
            envelope = None
        return envelope

    def format_value(self, value):
        """
        Return human readable string for value.

        It is the way ``value`` would be printed in REAPER GUI if it
        was the actual parameter value. Only works with FX that
        support Cockos VST extensions.

        Parameters
        ----------
        value : float
            Value to format.

        Returns
        -------
        formatted : str
            Formatted value.
        """
        parent_fx = self.parent_list.parent_fx
        parent = parent_fx.parent
        return self.functions["FormatParamValue"](
            parent.id, parent_fx.index, self.index, value, "", 2048
        )[5]

    @property
    def formatted(self):
        """
        Human readable string for parameter value.

        Only works with FX that support Cockos VST extensions.

        :type: str
        """
        parent_fx = self.parent_list.parent_fx
        parent = parent_fx.parent
        return self.functions["GetFormattedParamValue"](
            parent.id, parent_fx.index, self.index, "", 2048
        )[4]

    @property
    def name(self):
        """
        Parameter name.

        :type: str
        """
        parent_list = self.parent_list
        name = self.functions["GetParamName"](
            parent_list.parent_id, parent_list.fx_index, self.index, "", 2048
        )[4]
        return name

    @property
    def normalized(self):
        """
        Normalized FX parameter.

        Attribute can be set with a float, but be careful that since
        floats are immutable, this parameter won't have to right value
        anymore. See Examples below.

        :type: NormalizedFXParam

        Examples
        --------
        Say the parameter range is (0.0, 20.0).

        >>> param = fx.params[0]
        >>> param
        10.0
        >>> param.normalized
        0.5

        If you set the parameter like below, the parameter moves in
        REPAER, but the FXParam object you are using is not valid
        anymore.

        >>> param.normalized = 1
        >>> param, param.normalized
        10.0, 0.5

        You thus have to grab the updated FXParam from the FX like
        below.

        >>> param = fx.params[0]
        >>> param, param.normalized
        20.0, 1.0
        """
        min, max = self.range
        value = (self - min)/(max - min)
        return NormalizedFXParam(
            value, self.parent_list, self.index, self.functions
        )

    @normalized.setter
    def normalized(self, value):
        parent_fx = self.parent_list.parent_fx
        parent = parent_fx.parent
        self.functions["SetParamNormalized"](
            parent.id, parent_fx.id, self.index, value
        )

    @property
    def range(self):
        """
        Parameter range.

        :type: float, float
        """
        parent_list = self.parent_list
        min, max = self.functions["GetParam"](
            parent_list.parent_id, parent_list.fx_index, self.index, 0, 0
        )[-2:]
        return min, max


class FXParamsList(ReapyObjectList):

    """
    Container class for a list of FX parameters.

    Parameters can be accessed by name or index.

    Examples
    --------
    >>> params_list = fx.params
    >>> params_list[0]  # Say this is "Dry Gain" parameter
    0.5
    >>> params_list["Dry Gain"]
    0.5
    >>> params_list["Dry Gain"] = 0.1
    >>> params_list[0]
    0.1
    """

    def __init__(
        self, parent_fx=None, parent_id=None, parent_fx_index=None
    ):
        if parent_fx is None:
            parent_fx = reapy.FX(parent_id=parent_id, index=parent_fx_index)
        self.parent_id = parent_fx.parent_id
        self.fx_index = parent_fx.index
        self.functions = parent_fx.functions

    def __getitem__(self, i):
        with reapy.inside_reaper():
            if isinstance(i, str):
                i = self._get_param_index(i)
            n_params = len(self)
            if i >= n_params:
                raise IndexError(
                    "{} has only {} params".format(self.parent_fx, n_params)
                )
            i = i % n_params  # Allows for negative values
        value = self.functions["GetParam"](
            self.parent_id, self.fx_index, i, 0, 0
        )[0]
        param = FXParam(value, self, i, self.functions)
        return param

    def __iter__(self):
        for i, value in enumerate(self._get_values()):
            yield FXParam(value, self, i, self.functions)

    def __len__(self):
        length = self.parent_fx.n_params
        return length

    def __setitem__(self, i, value):
        with reapy.inside_reaper():
            if isinstance(i, str):
                i = self._get_param_index(i)
            n_params = len(self)
            if i >= n_params:
                raise IndexError(
                    "{} has only {} params".format(self.parent_fx, n_params)
                )
            i = i % n_params  # Allows for negative values
        self.functions["SetParam"](
            self.parent_id, self.fx_index, i, value
        )

    @reapy.inside_reaper()
    def _get_param_index(self, name):
        try:
            return [fx.name for fx in self].index(name)
        except ValueError:
            raise IndexError(
                "{} has no param named {}".format(self.parent_fx, name)
            )

    @reapy.inside_reaper()
    def _get_values(self):
        """Return values of all parameters in self."""
        return [
            self.functions["GetParam"](
                self.parent_id, self.fx_index, i, 0, 0
            )[0] for i in range(len(self))
        ]

    @property
    def _kwargs(self):
        return {
            "parent_fx_index": self.fx_index, "parent_id": self.parent_id
        }

    @property
    def parent_fx(self):
        """
        Parent FX.

        :type: FX
        """
        fx = reapy.FX(parent_id=self.parent_id, index=self.fx_index)
        return fx


class NormalizedFXParam(FXParam):

    """
    Normalized FX parameter.

    Access it via FXParam.normalized.

    Examples
    --------
    >>> fx.params[0]
    0.0
    >>> fx.params[0].range
    (-2.0, 0.0)
    >>> fx.params[0].normalized
    1.0
    >>> fx.params[0].normalized.range
    (0.0, 1.0)
    """

    def format_value(self, value):
        """
        Return human readable string for value.

        It is the way ``value`` would be printed in REAPER GUI if it
        was the actual parameter value. Only works with FX that
        support Cockos VST extensions.

        Parameters
        ----------
        value : float
            Value to format.

        Returns
        -------
        formatted : str
            Formatted value.
        """
        parent_fx = self.parent_list.parent_fx
        parent = parent_fx.parent
        return self.functions["FormatParamValueNormalized"](
            parent.id, parent_fx.index, self.index, value, "", 2048
        )[5]

    @property
    def range(self):
        """
        Parameter range (always equal to (0.0, 1.0)).
        """
        return (0.0, 1.0)

    @property
    def raw(self):
        """
        Raw (i.e. unnormalized) parameter.

        :type: FXParam
        """
        return self.parent_list[self.index]



from reapy import reascript_api as RPR
from reapy.core import ReapyObject
from reapy import Envelope
import typing as ty
from typing_extensions import TypedDict


class AutomationItem(ReapyObject):

    _class_name = "AutomationItem"
    envelope_id: int
    index: int

    def __init__(self,
                 envelope: ty.Optional[Envelope] = None,
                 index: int = 0,
                 envelope_id: ty.Optional[int] = None) -> None:
        ...

    @property
    def _kwargs(self) -> ty.Dict[str, int]:
        ...

    def delete_points_in_range(self, start: float, end: float) -> None:
        """
        Delete points between `start` and `end`.

        Parameters
        ----------
        start : float
            Range start in seconds.
        end : float
            Range end in seconds.
        """
        ...

    @property
    def length(self) -> float:
        """
        Return item length in seconds.

        Returns
        -------
        length : float
            Item length in seconds.
        """
        ...

    @length.setter
    def length(self, length: float) -> None:
        """
        Set item length.

        Parameters
        ----------
        length : float
            New item length in seconds.
        """
        ...

    @property
    def n_points(self) -> int:
        """
        Return number of automation points in item.

        Returns
        -------
        n_points : int
            Number of automation points in item.
        """
        ...

    @property
    def pool(self) -> int:
        """
        Return item pool.

        Returns
        -------
        pool : int
            Item pool.
        """
        ...

    @pool.setter
    def pool(self, pool: int) -> None:
        """
        Set item pool.

        Parameters
        ----------
        pool : int
            New item pool.
        """
        ...

    @property
    def position(self) -> float:
        """
        Return item position in seconds.

        Returns
        -------
        position : float
            Item position in seconds.
        """
        ...

    @position.setter
    def position(self, position: float) -> None:
        """
        Set item position.

        Parameters
        ----------
        position : float
            New item position in seconds.
        """
        ...



import reapy
import reapy.reascript_api as RPR
from reapy.core import ReapyObject, ReapyObjectList
from reapy.errors import DistError
import typing as ty


class FXParam(float):
    """FX parameter."""
    parent_list: FXParamsList
    index: int
    functions: ty.Dict[str, ty.Dict[str, ty.Callable[..., ty.Any]]]

    def __init__(
            self, value: float, parent_list: FXParamsList, index: int,
            functions: ty.Dict[str, ty.Dict[str, ty.Callable[..., ty.Any]]]
    ) -> None:
        ...

    def __new__(self, value: float, *args: ty.Any,
                **kwargs: ty.Any) -> FXParam:
        ...

    def add_envelope(self) -> reapy.Envelope:
        """
        Create envelope for the parameter and return it.

        Returns
        -------
        envelope : Envelope
            New envelope for the parameter.

        Notes
        -----
        If the envelope already exists, the function returns it.
        """
        ...

    @property
    def envelope(self) -> ty.Optional[reapy.Envelope]:
        """
        Parameter envelope (or None if it doesn't exist).

        :type: Envelope or NoneType
        """
        ...

    def format_value(self, value: float) -> str:
        """
        Return human readable string for value.

        It is the way ``value`` would be printed in REAPER GUI if it
        was the actual parameter value. Only works with FX that
        support Cockos VST extensions.

        Parameters
        ----------
        value : float
            Value to format.

        Returns
        -------
        formatted : str
            Formatted value.
        """
        ...

    @property
    def formatted(self) -> str:
        """
        Human readable string for parameter value.

        Only works with FX that support Cockos VST extensions.

        :type: str
        """
        ...

    @property
    def name(self) -> str:
        """
        Parameter name.

        :type: str
        """
        ...

    @property
    def normalized(self) -> 'NormalizedFXParam':
        """
        Normalized FX parameter.

        Attribute can be set with a float, but be careful that since
        floats are immutable, this parameter won't have to right value
        anymore. See Examples below.

        :type: NormalizedFXParam

        Examples
        --------
        Say the parameter range is (0.0, 20.0).

        >>> param = fx.params[0]
        >>> param
        10.0
        >>> param.normalized
        0.5

        If you set the parameter like below, the parameter moves in
        REPAER, but the FXParam object you are using is not valid
        anymore.

        >>> param.normalized = 1
        >>> param, param.normalized
        10.0, 0.5

        You thus have to grab the updated FXParam from the FX like
        below.

        >>> param = fx.params[0]
        >>> param, param.normalized
        20.0, 1.0
        """
        ...

    @normalized.setter
    def normalized(self, value: float) -> None:
        ...

    @property
    def range(self) -> ty.Tuple[float, float]:
        """
        Parameter range.

        :type: float, float
        """
        ...

    def __eq__(self, other: object) -> bool: ...

    def __ne__(self, other: object) -> bool: ...

    def __gt__(self, other: object) -> bool: ...

    def __ge__(self, other: object) -> bool: ...

    def __lt__(self, other: object) -> bool: ...

    def __le__(self, other: object) -> bool: ...


class FXParamsList(ReapyObjectList):
    """
    Container class for a list of FX parameters.

    Parameters can be accessed by name or index.

    Examples
    --------
    >>> params_list = fx.params
    >>> params_list[0]  # Say this is "Dry Gain" parameter
    0.5
    >>> params_list["Dry Gain"]
    0.5
    >>> params_list["Dry Gain"] = 0.1
    >>> params_list[0]
    0.1
    """
    parent_id: str
    fx_index: int
    functions: ty.Dict[str, ty.Callable[..., ty.Any]]

    def __init__(self,
                 parent_fx: ty.Optional[reapy.FX] = None,
                 parent_id: ty.Optional[str] = None,
                 parent_fx_index: ty.Optional[int] = None) -> None:
        ...

    def __getitem__(self, i: ty.Union[str, int]) -> FXParam:
        ...

    def __iter__(self) -> ty.Iterator[FXParam]:
        ...

    def __len__(self) -> int:
        ...

    def __setitem__(self, i: ty.Union[str, int], value: float) -> None:
        ...

    @reapy.inside_reaper()
    def _get_param_index(self, name: str) -> int:
        ...

    @reapy.inside_reaper()
    def _get_values(self) -> ty.List[float]:
        """Return values of all parameters in self."""
        ...

    @property
    def _kwargs(self) -> ty.Dict[str, ty.Union[int, str]]:
        ...

    @property
    def parent_fx(self) -> reapy.FX:
        """
        Parent FX.

        :type: FX
        """
        ...


class NormalizedFXParam(FXParam):
    """
    Normalized FX parameter.

    Access it via FXParam.normalized.

    Examples
    --------
    >>> fx.params[0]
    0.0
    >>> fx.params[0].range
    (-2.0, 0.0)
    >>> fx.params[0].normalized
    1.0
    >>> fx.params[0].normalized.range
    (0.0, 1.0)
    """

    def format_value(self, value: float) -> str:
        """
        Return human readable string for value.

        It is the way ``value`` would be printed in REAPER GUI if it
        was the actual parameter value. Only works with FX that
        support Cockos VST extensions.

        Parameters
        ----------
        value : float
            Value to format.

        Returns
        -------
        formatted : str
            Formatted value.
        """
        ...

    @property
    def range(self) -> ty.Tuple[float, float]:
        """
        Parameter range (always equal to (0.0, 1.0)).
        """
        ...

    @property
    def raw(self) -> FXParam:
        """
        Raw (i.e. unnormalized) parameter.

        :type: FXParam
        """
        ...



from .reapy_object import ReapyObject, ReapyObjectList

from .audio_accessor import AudioAccessor
from .envelope import Envelope, EnvelopeList
from .fx import FX, FXList, FXParam, FXParamsList
from .item import CC, CCList, Item, Note, NoteList, Source, Take
from .map import map
from .project import Marker, Project, Region, TimeSelection
from .track import AutomationItem, Send, Track, TrackList
from .window import MIDIEditor, ToolTip, Window


__all__ = [
    # core.reapy_object
    "ReapyObject",
    "ReapyObjectList",
    # core.audio_accessor
    "AudioAccessor",
    # core.envelope
    "Envelope",
    "EnvelopeList",
    # core.fx
    "FX",
    "FXList",
    "FXParam",
    "FXParamsList",
    # core.item
    "CC",
    "CCList",
    "Item",
    "Note",
    "NoteList",
    "Source",
    "Take",
    # core.map
    "map",
    # core.project
    "Marker",
    "Project",
    "Region",
    "TimeSelection",
    # core.track
    "AutomationItem",
    "Send",
    "Track",
    "TrackList",
    # core.window
    "MIDIEditor",
    "ToolTip",
    "Window",
]



"""Define FX and FXParam classes."""

import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject, ReapyObjectList
from reapy.errors import DistError, UndefinedFXParamError


class FX(ReapyObject):

    """FX on a Track or a Take."""

    _class_name = "FX"
    functions = {
        prefix: {
            name.replace(prefix, ""): function
            for name, function in RPR.__dict__.items()
            if name.startswith(prefix)
        }
        for prefix in ("TrackFX_", "TakeFX_")
    }

    def __init__(self, parent=None, index=None, parent_id=None):
        if parent_id is None:
            message = (
                "One of `parent` or `parent_id` must be specified."
            )
            assert parent is not None, message
            parent_id = parent.id
        self.parent_id = parent_id
        self.index = index
        self.functions = self._get_functions()

    def _get_functions(self):
        if isinstance(self.parent, reapy.Track):
            type = "TrackFX_"
        else:
            type = "TakeFX_"
        return self.functions[type]

    @property
    def _kwargs(self):
        return {"parent_id": self.parent_id, "index": self.index}

    def close_chain(self):
        """Close FX chain."""
        self.functions["Show"](self.parent.id, self.index, 0)

    def close_floating_window(self):
        """Close FX floating window."""
        self.functions["Show"](self.parent.id, self.index, 2)

    def close_ui(self):
        """Close user interface."""
        self.is_ui_open = False

    def copy_to_take(self, take, index=0):
        """
        Copy FX to take.

        Parameters
        ----------
        take : Take
            Destination take.
        index : int
            Index on destination take.

        See also
        --------
        FX.move_to_take
        """
        self.functions["CopyToTake"](
            self.parent_id, self.index, take.id, index, False
        )

    def copy_to_track(self, track, index=0):
        """
        Copy FX to track.

        Parameters
        ----------
        track : Track
            Destination track.
        index : int
            Index on destination track.

        See also
        --------
        FX.move_to_track
        """
        self.functions["CopyToTrack"](
            self.parent_id, self.index, track.id, index, False
        )

    def delete(self):
        """Delete FX."""
        self.functions["Delete"](self.parent_id, self.index)

    def disable(self):
        """Disable FX."""
        self.is_enabled = False

    def enable(self):
        """Enable FX."""
        self.is_enabled = True

    @property
    def is_enabled(self):
        """
        Whether FX is enabled.

        :type: bool
        """
        is_enabled = bool(
            self.functions["GetEnabled"](self.parent_id, self.index)
        )
        return is_enabled

    @is_enabled.setter
    def is_enabled(self, enabled):
        self.functions["SetEnabled"](self.parent_id, self.index, enabled)

    @property
    def is_online(self):
        """
        Whether FX is online.

        :type: bool
        """
        is_online = not bool(
            self.functions["GetOffline"](self.parent_id, self.index)
        )
        return is_online

    @is_online.setter
    def is_online(self, online):
        offline = not online
        self.functions["SetOffline"](self.parent_id, self.index, offline)

    @property
    def is_ui_open(self):
        """
        Whether FX user interface is open.

        :type: bool
        """
        is_ui_open = bool(
            self.functions["GetOpen"](self.parent_id, self.index)
        )
        return is_ui_open

    @is_ui_open.setter
    def is_ui_open(self, open):
        self.functions["SetOpen"](self.parent_id, self.index, open)

    def make_offline(self):
        """Make FX offline."""
        self.is_online = False

    def make_online(self):
        """Make FX online."""
        self.is_online = True

    def move_to_take(self, take, index=0):
        """
        Move FX to take.

        Parameters
        ----------
        take : Take
            Destination take.
        index : int
            Index on destination take.

        See also
        --------
        FX.copy_to_take
        """
        self.functions["CopyToTake"](
            self.parent_id, self.index, take.id, index, True
        )

    def move_to_track(self, track, index=0):
        """
        Move FX to track.

        Parameters
        ----------
        track : Track
            Destination track.
        index : int
            Index on destination track.

        See also
        --------
        FX.copy_to_track
        """
        self.functions["CopyToTrack"](
            self.parent_id, self.index, track.id, index, True
        )

    @property
    def n_inputs(self):
        """
        Number of inputs of FX.

        :type: int
        """
        return self.functions["GetIOSize"](
            self.parent.id, self.index, 0, 0
        )[3]

    @property
    def n_outputs(self):
        """
        Number of outputs of FX.

        :type: int
        """
        return self.functions["GetIOSize"](
            self.parent.id, self.index, 0, 0
        )[4]

    @property
    def n_params(self):
        """
        Number of parameters.

        :type: int
        """
        n_params = self.functions["GetNumParams"](self.parent_id, self.index)
        return n_params

    @property
    def n_presets(self):
        """
        Number of presets.

        :type: int
        """
        n_presets = self.functions["GetPresetIndex"](
            self.parent_id, self.index, 0
        )[-1]
        return n_presets

    @property
    def name(self):
        """
        FX name.

        :type: str
        """
        name = self.functions["GetFXName"](
            self.parent_id, self.index, "", 2048
        )[3]
        return name

    def open_chain(self):
        """Open FX chain with focus on FX."""
        self.functions["Show"](self.parent.id, self.index, 1)

    def open_floating_window(self):
        """Open FX floating window."""
        self.functions["Show"](self.parent.id, self.index, 3)

    def open_ui(self):
        """Open FX user interface."""
        self.is_ui_open = True

    @property
    def params(self):
        """
        List of parameters.

        :type: FXParamsList
        """
        params = reapy.FXParamsList(self)
        return params

    @property
    def parent(self):
        """
        FX parent.

        :type: Track or Take
        """
        if self.parent_id.startswith("(MediaTrack*)"):
            return reapy.Track(self.parent_id)
        return reapy.Take(self.parent_id)

    @property
    def preset(self):
        """
        FX preset name.

        :type: str

        Attribute can be set by passing a str or int. In the first
        case, the str can either be a preset name or the path to a
        .vstpreset file. Otherwise, the int is the preset index.
        """
        preset = self.functions["GetPreset"](
            self.parent_id, self.index, "", 2048
        )[3]
        return preset

    @preset.setter
    def preset(self, preset):
        """
        Set FX preset.

        Parameters
        ----------
        preset : str or int
            If str, preset name or path to .vstpreset file. If int,
            preset index. Set to -2 for factory preset, and -1 for user
            default preset.
        """
        if isinstance(preset, str):
            self.functions["SetPreset"](
                self.parent_id, self.index, preset
            )
        elif isinstance(preset, int):
            self.functions["SetPresetByIndex"](
                self.parent_id, self.index, preset
            )

    @property
    def preset_index(self):
        """
        FX preset index.

        :type: int
        """
        index = self.functions["GetPresetIndex"](
            self.parent_id, self.index, 0
        )[0]
        return index

    @property
    def preset_file(self):
        """
        Path to FX preset file.

        :type: str
        """
        file = self.functions["GetUserPresetFilename"](
            self.parent_id, self.index, "", 2048
        )[2]
        return file

    def use_previous_preset(self):
        """Use previous preset in the presets list."""
        self.functions["NavigatePresets"](self.parent_id, self.index, -1)

    def use_next_preset(self):
        """Use next preset in the presets list."""
        self.functions["NavigatePresets"](self.parent_id, self.index, 1)

    @property
    def window(self):
        """
        Floating window associated to FX, if it exists.

        :type: Window or NoneType
        """
        window = reapy.Window(
            self.functions["GetFloatingWindow"](self.parent.id, self.index)
        )
        if not window._is_defined:
            window = None
        return window


class FXList(ReapyObjectList):

    """
    Container class for a list of FXs.

    FXs can be accessed by name or index.

    Examples
    --------
    >>> fx_list = track.fxs
    >>> fx_list[0]
    FX(parent_id="(MediaTrack*)0x0000000006CDEBE0", index=0)
    >>> len(fx_list)
    1
    >>> fx_list["VST: ReaComp (Cockos)"]
    FX(parent_id="(MediaTrack*)0x0000000006CDEBE0", index=0)
    """

    _class_name = "FXList"

    def __init__(self, parent):
        self.parent = parent

    @reapy.inside_reaper()
    def __delitem__(self, key):
        fxs = self[key] if isinstance(key, slice) else [self[key]]
        for fx in fxs:
            fx.delete()

    def __getitem__(self, i):
        if isinstance(i, slice):
            return self._get_items_from_slice(i)
        with reapy.inside_reaper():
            if isinstance(i, str):
                i = self._get_fx_index(name=i)
            n_fxs = self.parent.n_fxs
        if i >= n_fxs:
            raise IndexError("{} has only {} fxs".format(self.parent, n_fxs))
        i = i % n_fxs  # Allows for negative values
        fx = FX(self.parent, i)
        return fx

    def __len__(self):
        return self.parent.n_fxs

    @reapy.inside_reaper()
    def _get_items_from_slice(self, slice):
        indices = range(*slice.indices(len(self)))
        return [self[i] for i in indices]

    def _get_fx_index(self, name):
        name = name[name.find(': ') + 2:]  # Remove FX type prefix
        if isinstance(self.parent, reapy.Track):
            prefix = "TrackFX_"
            args = (self.parent.id, name, False, 0)
        else:
            prefix = "TakeFX_"
            args = (self.parent.id, name, 0)
        index = getattr(RPR, prefix + "AddByName")(*args)
        if index == -1:
            raise KeyError("No FX named {}".format(name))
        return index

    @property
    def _args(self):
        return self.parent,



"""
Build ``docs/source/api_table.rst`` from ``docs/api.json``.
"""

import reapy
import reapy.core

from collections import defaultdict
import inspect
import json
import os
from sphinx.cmd.build import build_main as sphinx_build_cmd


HEADER = """
Translation Table
=================

On this page you can find the list of all ReaScript API functions with their ``reapy`` counterparts (when they exist).

Categories are based on Mespotine's `ReaScript documentation <https://mespotin.uber.space/Mespotine/Ultraschall/Reaper_Api_Documentation.html>`_.

.. contents:: Categories
    :local:
    :depth: 2


"""


def build_api_table():
    api_infos = load_api_infos()
    groups = get_groups(api_infos)
    output_path = get_output_path()
    f = open(output_path, "w")
    f.write(HEADER)
    for group_name, group in sorted(groups.items()):
        f.write(get_group_string(group_name, group))
    f.close()


def double_quoted(s):
    return repr(s).replace("'", '"')


def load_api_infos():
    path = get_input_path()
    with open(path) as f:
        return json.load(f)


def get_changelog_links():
    api = load_api_infos()
    names = set(
        n for info in api.values() for n in info["reapy"] if n != "DEPRECATED"
    )
    links = sorted([reapy_link(n, type="md") + "\n" for n in names])
    links.append(
        "[`inside_reaper`]: (https://python-reapy.readthedocs.io/en/"
        "latest/reapy.tools.html#reapy.tools.inside_reaper.inside_reaper)"
    )
    return links


def get_group_contents(group):
    string = (
        ".. csv-table::\n"
        "\t:header: \"ReaScript API function\", \"reapy API function\""
        "\n\n"
    )
    for reascript_name, infos in group.items():
        reapy_names = infos["reapy"]
        string += "\t{},{}\n".format(
            double_quoted(reascript_link(reascript_name)),
            double_quoted("; ".join(map(reapy_link, reapy_names)))
        )
    return string


def get_group_string(group_name, group):
    contents = get_group_contents(group)
    return "{}\n{}\n\n{}\n\n".format(
        group_name, "-"*len(group_name), contents
    )


def get_groups(infos):
    groups = defaultdict(dict)
    for name, info in infos.items():
        groups[info["group"]][name] = info
    return groups


def get_input_path():
    path = os.path.join(os.path.dirname(__file__), "docs", "api.json")
    return path


def get_output_path():
    output_path = os.path.join(
        os.path.dirname(__file__), "docs", "source", "api_table.rst"
    )
    return output_path


def get_reapy_module(s):
    parts = s.split(".")
    if s[0].lower() != s[0]:
        parts = parts[:-1]
    o = reapy
    for a in parts:
        o = getattr(o, a)
    module = inspect.getmodule(o)
    return module


def markdown_link(s, url):
    return "[{}]: {}".format(s, url)


def reapy_link(s, type="rst"):
    if s == "DEPRECATED":
        return s
    if s in ("defer", "at_exit"):
        url = "reapy.core.reaper.html#reapy.core.reaper.defer.{}".format(s)
    else:
        if s[0].lower() != s[0]:
            page = "reapy.core.html"
            anchor = "reapy.core." + s
        else:
            page = "reapy.core.reaper.html"
            anchor = "reapy.core.reaper."
            if "." in s:
                anchor += s
            else:
                anchor += "reaper." + s
        url = page + "#" + anchor
    if type == "rst":
        return rst_link(s, url)
    else:
        url = "https://python-reapy.readthedocs.io/en/latest/" + url
        return markdown_link("`{}`".format(s), url)


def reascript_link(s):
    page = "https://www.reaper.fm/sdk/reascript/reascripthelp.html"
    anchor = s
    if s in ("defer", "atexit"):
        anchor = "python_" + anchor
    url = page + "#" + anchor
    return rst_link(s, url)


def rst_link(s, url):
    return "`{} <{}>`_".format(s, url)


def sphinx_build():
    root = os.path.dirname(__file__)
    source = os.path.join(root, "docs", "source")
    build = os.path.join(root, "docs", "build", "html")
    sphinx_build_cmd([source, build])


def update_changelog():
    changelog_path = os.path.join(os.path.dirname(__file__), "CHANGELOG.md")
    with open(changelog_path) as f:
        lines = f.readlines()
    log = lines[:lines.index("[//]: # (LINKS)\n") + 1]
    with open(changelog_path, "w") as f:
        f.writelines(log)
        f.writelines(get_changelog_links())


if __name__ == "__main__":
    build_api_table()
    update_changelog()
    sphinx_build()



import reapy
import reapy.reascript_api as RPR


class ReapyMetaclass(type):

    @property
    def _reapy_parent(self):
        """Return first reapy parent class."""
        for candidate in self.__mro__:
            if candidate.__module__.startswith('reapy.'):
                return candidate


class ReapyObject(metaclass=ReapyMetaclass):
    """Base class for reapy objects."""

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __repr__(self):

        def to_str(x):
            if isinstance(x, str):
                return "\"{}\"".format(x)
            return str(x)

        args = ", ".join(map(to_str, self._args))
        kwargs = ", ".join(
            ("{}={}".format(k, to_str(v)) for k, v in self._kwargs.items())
        )
        if args and kwargs:
            brackets = ", ".join((args, kwargs))
        else:
            brackets = args if args else kwargs
        rep = "{}({})".format(self.__class__.__name__, brackets)
        return rep

    @property
    def _args(self):
        return ()

    def _get_pointer_and_name(self):
        name, pointer = self.id.split(')')
        return int(pointer, base=16), name[1:]

    @property
    def _is_defined(self):
        if hasattr(self, "id"):
            return not self.id.endswith("0x0000000000000000")
        raise NotImplementedError

    @property
    def _kwargs(self):
        return {}

    def _to_dict(self):
        return {
            "__reapy__": True,
            "class": self.__class__._reapy_parent.__name__,
            "args": self._args,
            "kwargs": self._kwargs
        }


class ReapyObjectList(ReapyObject):
    """Abstract class for list of ReapyObjects."""

    pass



"""Define tools such as Program and custom json module."""

import reapy
from ._inside_reaper import inside_reaper, dist_api_is_enabled
from .network.machines import connect, connect_to_default_machine, reconnect
from .extension_dependency import depends_on_sws, depends_on_extension

__all__ = [
    'inside_reaper',
    'dist_api_is_enabled',
    'connect',
    'connect_to_default_machine',
    'reconnect',
    'depends_on_sws',
    'depends_on_extension',
]



import reapy
import reapy.reascript_api as RPR
import typing as ty


def get_active_editor() -> ty.Optional[reapy.MIDIEditor]:
    """
    Return active MIDI editor, or None if no editor is active.

    Returns
    -------
    editor : MIDIEditor or None
        Active MIDI editor, or None if no editor is active.
    """
    ...


@reapy.inside_reaper()
def get_input_names() -> ty.List[str]:
    """
    Return names of all input channels.

    Returns
    -------
    names : list of str
        Names of input channels.
    """
    ...


def get_max_inputs() -> int:
    """
    Return maximum number of MIDI inputs.

    Returns
    -------
    max_inputs : int
        Maximum number of MIDI inputs.
    """
    ...


def get_max_outputs() -> int:
    """
    Return maximum number of MIDI outputs.

    Returns
    -------
    max_outputs : int
        Maximum number of MIDI outputs.
    """
    ...


def get_n_inputs() -> int:
    """
    Return number of MIDI inputs.

    Returns
    -------
    n_inputs : int
        Number of MIDI inputs.
    """
    ...


def get_n_outputs() -> int:
    """
    Return number of MIDI outputs.

    Returns
    -------
    n_outputs : int
        Number of MIDI outputs.
    """
    ...


@reapy.inside_reaper()
def get_output_names() -> ty.List[str]:
    """
    Return names of all output channels.

    Returns
    -------
    names : list of str
        Names of output channels.
    """
    ...


def reinit() -> None:
    """Reset all MIDI devices."""
    ...



import warnings

import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject


class Envelope(ReapyObject):

    def __init__(self, parent, id):
        self.id = id
        self._parent = parent

    @property
    def _args(self):
        return (self.parent, self.id)

    def add_item(self, position=0., length=1., pool=0):
        """
        Add automation item to envelope.

        Parameters
        ----------
        position : float, optional
            New item position in seconds (default=0).
        length : float
            New item length in seconds (default=0).
        pool : int
            New item pool index. If >= 0 the automation item will be a
            new instance of that pool (which will be created as an
            empty instance if it does not exist).

        Returns
        -------
        item : reapy.AutomationItem
            New automation item.
        """
        item_index = RPR.InsertAutomationItem(self.id, pool, position, length)
        item = reapy.AutomationItem(envelope=self, index=item_index)
        return item

    def delete_points_in_range(self, start, end):
        """
        Delete envelope points between `start` and `end`.

        Parameters
        ----------
        start : float
            Range start in seconds.
        end : float
            Range end in seconds.
        """
        RPR.DeleteEnvelopePointRange(self.id, start, end)

    @reapy.inside_reaper()
    def get_derivatives(self, time, raw=False):
        """
        Return envelope derivatives of order 1, 2, 3 at a given time.

        Parameters
        ----------
        time : float
            Time in seconds.
        raw : bool, optional
            Whether to return raw values or the human-readable version
            which is printed in REAPER GUI (default=False).

        Returns
        -------
        d, d2, d3 : float
            First, second and third order derivatives.

        Examples
        --------
        >>> envelope = track.envelopes["Pan"]
        >>> envelope.get_derivatives(10, raw=True)
        (0.10635556358181712, 0.2127113398749741, 0.21271155258652666)
        >>> envelope.get_value(10)  # human-readable
        ('10%L', '21%L', '21%L')
        """
        d, d2, d3 = RPR.Envelope_Evaluate(self.id, time, 1, 1, 0, 0, 0, 0)[6:]
        if not raw:
            d = RPR.Envelope_FormatValue(self.id, d, "", 2048)[2]
            d2 = RPR.Envelope_FormatValue(self.id, d2, "", 2048)[2]
            d3 = RPR.Envelope_FormatValue(self.id, d3, "", 2048)[2]
        return d, d2, d3

    @reapy.inside_reaper()
    def get_value(self, time, raw=False):
        """
        Return envelope value at a given time.

        Parameters
        ----------
        time : float
            Time in seconds.
        raw : bool, optional
            Whether to return raw value or its human-readable version,
            which is the one that is printed in REAPER GUI
            (default=False).

        Returns
        -------
        value : float or str
            Envelope value.

        Examples
        --------
        >>> envelope = track.envelopes["Pan"]
        >>> envelope.get_value(10, raw=True)
        -0.5145481809245827
        >>> envelope.get_value(10)  # human-readable
        '51%R'
        """
        value = RPR.Envelope_Evaluate(self.id, time, 0, 0, 0, 0, 0, 0)[5]
        if not raw:
            value = RPR.Envelope_FormatValue(self.id, value, "", 2048)[2]
        return value

    @reapy.inside_reaper()
    @property
    def has_valid_id(self):
        """
        Whether ReaScript ID is still valid.

        For instance, if envelope has been deleted, ID will not be valid
        anymore.

        :type: bool
        """
        try:
            project_id = self.parent.project.id
        except (OSError, AttributeError):
            return False
        pointer, name = self._get_pointer_and_name()
        return bool(RPR.ValidatePtr2(project_id, pointer, name))

    @property
    def items(self):
        """
        List of automation items in envelope.

        :type: list of reapy.AutomationItem
        """
        items = [reapy.AutomationItem(self, i) for i in range(self.n_items)]
        return items

    @property
    def n_items(self):
        """
        Number of automation items in envelope.

        :type: int
        """
        n_items = RPR.CountAutomationItems(self.id)
        return n_items

    @property
    def n_points(self):
        """
        Number of points in envelope.

        :type: int
        """
        n_points = RPR.CountEnvelopePoints(self.id)
        return n_points

    @property
    def name(self):
        """
        Envelope name.

        :type: str
        """
        name = RPR.GetEnvelopeName(self.id, "", 2048)[2]
        return name

    @property
    def parent(self):
        """
        Envelope parent.

        :type: Take or Track
        """
        return self._parent


class EnvelopeList(ReapyObject):

    """
    Container class for the list of envelopes on a Take or Track.

    Envelopes can be accessed from the EnvelopeList either by index,
    name or chunk_name (e.g. "<VOLENV").

    Examples
    --------
    >>> len(track.envelopes)
    2
    >>> envelope = track.envelopes[0]
    >>> envelope.name
    'Volume'
    >>> envelope == track.envelopes["Volume"]
    True
    >>> envelope == track.envelopes["<VOLENV"]
    True
    >>> [e.name for e in track.envelopes]
    ['Volume', 'Pan']
    """

    def __init__(self, parent):
        self.parent = parent

    @property
    def _args(self):
        return (self.parent,)

    def __getitem__(self, key):
        parent_type = self.parent.__class__._reapy_parent.__name__
        attr = "Get{}Envelope".format(parent_type)
        if isinstance(key, str):
            if key.startswith("<") and parent_type == 'Track':
                attr += "ByChunkName"
            else:
                attr += "ByName"
        callback = getattr(RPR, attr)
        envelope = Envelope(self.parent, callback(self.parent.id, key))
        if not envelope._is_defined:
            raise KeyError("No envelope for key {}".format(repr(key)))
        return envelope

    def __len__(self):
        return self.parent.n_envelopes



import socket


class Socket:

    """
    Wrapped `socket` that can send and receive data of any length.
    """

    def __init__(self, s=None):
        self._socket = socket.socket() if s is None else s
        self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    @staticmethod
    def _non_blocking(f):
        """
        Modify a socket method so that it returns `None` when time
        out is reached.
        """
        def g(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except socket.timeout:
                pass
        return g

    def accept(self, *args, **kwargs):
        connection, address = self._socket.accept(*args, **kwargs)
        connection = Socket(connection)
        return connection, address

    def bind(self, *args, **kwargs):
        return self._socket.bind(*args, **kwargs)

    def close(self, *args, **kwargs):
        return self._socket.close(*args, **kwargs)

    def connect(self, *args, **kwargs):
        return self._socket.connect(*args, **kwargs)

    def listen(self, *args, **kwargs):
        return self._socket.listen(*args, **kwargs)

    def recv(self, timeout=.0001):
        """Receive data of arbitrary length."""
        # First get data length
        self.settimeout(timeout)
        length = self._socket.recv(8)
        length = int.from_bytes(length, "little")
        if length == 0:
            raise ConnectionAbortedError
        # Then receive data
        self.settimeout(None)
        return self._socket.recv(length, socket.MSG_WAITALL)

    def send(self, data):
        """Send data."""
        # First send data length
        length = len(data).to_bytes(8, "little")
        self._socket.sendall(length)
        # Then send data
        self._socket.sendall(data)

    def settimeout(self, *args, **kwargs):
        return self._socket.settimeout(*args, **kwargs)

    def shutdown(self, *args, **kwargs):
        return self._socket.shutdown(*args, **kwargs)



"""
Activate ``reapy`` server.

Running this ReaScript from inside REAPER sets the ``reapy`` server
that receives and executes API calls requests from outside. It will
automatically be run when importing ``reapy`` from outside, if it is
enabled.
"""

import reapy

import os
import site

from reapy.tools.network import Server

SERVER: Server


def run_main_loop() -> None:
    ...


def get_new_reapy_server() -> Server:
    ...
